import logging

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.hermes.models import CGDeliverables
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.rnafusion.base import rnafusion, start, start_available, store, store_available
from cg.constants import EXIT_SUCCESS
from cg.constants.tb import AnalysisStatus
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig


def test_rnafusion_no_args(cli_runner: CliRunner, rnafusion_context: CGConfig):
    """Test to see that running BALSAMIC without options prints help and doesn't result in an error."""
    # GIVEN no arguments or options besides the command call

    # WHEN running command
    result = cli_runner.invoke(rnafusion, [], obj=rnafusion_context)

    # THEN command runs successfully
    print(result.output)

    assert result.exit_code == EXIT_SUCCESS

    # THEN help should be printed
    assert "help" in result.output


def test_start(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    rnafusion_case_id: str,
    mock_config,
):
    """Test to ensure all parts of start command will run successfully given ideal conditions."""
    caplog.set_level(logging.INFO)

    # GIVEN case id
    case_id: str = rnafusion_case_id

    # GIVEN a mocked config

    # GIVEN decompression is not needed
    RnafusionAnalysisAPI.resolve_decompression.return_value = None

    # WHEN dry running with dry specified
    result = cli_runner.invoke(start, [case_id, "--dry-run"], obj=rnafusion_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    assert case_id in caplog.text

    # THEN command should not include resume flag
    assert "-resume" not in caplog.text


def test_store_success(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    real_housekeeper_api: HousekeeperAPI,
    mock_deliverable,
    mock_analysis_finish,
    caplog: LogCaptureFixture,
    hermes_deliverables: dict,
    mocker,
    rnafusion_case_id: str,
):
    """Test to ensure all parts of store command are run successfully given ideal conditions."""
    caplog.set_level(logging.INFO)

    # GIVEN case-id for which we created a config file, deliverables file, and analysis_finish file
    case_id: str = rnafusion_case_id

    # Set Housekeeper to an empty real Housekeeper store
    rnafusion_context.housekeeper_api_: HousekeeperAPI = real_housekeeper_api
    rnafusion_context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # Make sure the bundle was not present in hk
    assert not rnafusion_context.housekeeper_api.bundle(case_id)

    # Make sure analysis not already stored in status_db
    assert not rnafusion_context.status_db.get_case_by_internal_id(internal_id=case_id).analyses

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # WHEN running command
    result = cli_runner.invoke(store, [case_id], obj=rnafusion_context)
    # THEN bundle should be successfully added to HK and STATUSDB
    assert result.exit_code == EXIT_SUCCESS
    assert "Analysis successfully stored in Housekeeper" in caplog.text
    assert "Analysis successfully stored in StatusDB" in caplog.text
    assert rnafusion_context.status_db.get_case_by_internal_id(internal_id=case_id).analyses
    assert rnafusion_context.housekeeper_api.bundle(case_id)


def test_store_fail(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    real_housekeeper_api: HousekeeperAPI,
    mock_deliverable,
    mock_analysis_finish,
    caplog: LogCaptureFixture,
    hermes_deliverables: dict,
    mocker,
):
    """Test store command fails when a case did not finish."""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID where analysis finish is not mocked
    case_id_fail: str = "rnafusion_case_not_finished"

    # Set Housekeeper to an empty real Housekeeper store
    rnafusion_context.housekeeper_api_: HousekeeperAPI = real_housekeeper_api
    rnafusion_context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # Make sure the bundle was not present in hk
    assert not rnafusion_context.housekeeper_api.bundle(case_id_fail)

    # WHEN running command
    result_fail = cli_runner.invoke(store, [case_id_fail], obj=rnafusion_context)

    # THEN bundle exist status should be
    assert result_fail.exit_code != EXIT_SUCCESS


def test_start_available(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    mocker,
    rnafusion_case_id: str,
    mock_config,
):
    """Test to ensure all parts of compound start-available command are executed given ideal conditions
    Test that start-available picks up eligible cases and does not pick up ineligible ones."""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID of sample where read counts pass threshold
    case_id_success: str = rnafusion_case_id

    # GIVEN a mocked config

    # GIVEN decompression is not needed
    mocker.patch.object(RnafusionAnalysisAPI, "resolve_decompression")
    RnafusionAnalysisAPI.resolve_decompression.return_value = None

    # WHEN running command
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=rnafusion_context)

    # THEN command exits with 0
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should successfully identify the one case eligible for auto-start
    assert case_id_success in caplog.text


def test_store_available(
    tmpdir_factory,
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    real_housekeeper_api,
    mock_deliverable,
    mock_analysis_finish,
    caplog: LogCaptureFixture,
    mocker,
    hermes_deliverables,
    rnafusion_case_id: str,
    mock_config,
):
    """Test to ensure all parts of compound store-available command are executed given ideal conditions
    Test that sore-available picks up eligible cases and does not pick up ineligible ones."""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID of sample where read counts pass threshold
    case_id_success: str = rnafusion_case_id

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # GIVEN a mocked config

    # Ensure case was successfully picked up by start-available and status set to running
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=rnafusion_context)
    rnafusion_context.status_db.get_case_by_internal_id(case_id_success).action = "running"
    rnafusion_context.status_db.session.commit()

    # THEN command exits with 0
    assert result.exit_code == EXIT_SUCCESS
    assert case_id_success in caplog.text
    assert rnafusion_context.status_db.get_case_by_internal_id(case_id_success).action == "running"

    rnafusion_context.housekeeper_api_ = real_housekeeper_api
    rnafusion_context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # WHEN running command
    result = cli_runner.invoke(store_available, obj=rnafusion_context)

    # THEN command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN case id with analysis_finish gets picked up
    assert case_id_success in caplog.text

    # THEN case has analyses
    assert rnafusion_context.status_db.get_case_by_internal_id(case_id_success).analyses

    # THEN bundle can be found in Housekeeper
    assert rnafusion_context.housekeeper_api.bundle(case_id_success)

    # THEN bundle added successfully and action set to None
    assert rnafusion_context.status_db.get_case_by_internal_id(case_id_success).action is None



"""CLI support to create config and/or start BALSAMIC """

import logging

import click

from cg.cli.workflow.balsamic.base import (
    config_case,
    run,
    report_deliver,
    store_housekeeper,
    start,
    start_available,
    store,
    store_available,
)
from cg.cli.workflow.commands import link, resolve_compression
from cg.meta.workflow.balsamic_qc import BalsamicQCAnalysisAPI

LOG = logging.getLogger(__name__)


@click.group("balsamic-qc", invoke_without_command=True)
@click.pass_context
def balsamic_qc(context: click.Context):
    """Cancer analysis workflow"""
    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None
    config = context.obj
    context.obj.meta_apis["analysis_api"] = BalsamicQCAnalysisAPI(
        config=config,
    )


balsamic_qc.add_command(resolve_compression)
balsamic_qc.add_command(link)
balsamic_qc.add_command(config_case)
balsamic_qc.add_command(run)
balsamic_qc.add_command(report_deliver)
balsamic_qc.add_command(store_housekeeper)
balsamic_qc.add_command(start)
balsamic_qc.add_command(start_available)
balsamic_qc.add_command(store)
balsamic_qc.add_command(store_available)



""" Trailblazer API for cg """ ""
import datetime
import datetime as dt
import logging
from typing import Any, Optional

from google.auth import jwt
from google.auth.crypt import RSASigner

from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import Pipeline
from cg.constants.constants import APIMethods, FileFormat, WorkflowManager
from cg.constants.priority import SlurmQos
from cg.constants.tb import AnalysisStatus
from cg.exc import TrailblazerAPIHTTPError
from cg.io.controller import APIRequest, ReadStream

LOG = logging.getLogger(__name__)


class TrailblazerAPI:
    """Interface to Trailblazer for `cg`."""

    __STARTED_STATUSES = [
        AnalysisStatus.COMPLETED,
        AnalysisStatus.FAILED,
        AnalysisStatus.PENDING,
        AnalysisStatus.RUNNING,
        AnalysisStatus.ERROR,
        AnalysisStatus.QC,
    ]
    __ONGOING_STATUSES = [
        AnalysisStatus.PENDING,
        AnalysisStatus.RUNNING,
        AnalysisStatus.ERROR,
        AnalysisStatus.QC,
    ]

    def __init__(self, config: dict):
        self.service_account = config["trailblazer"]["service_account"]
        self.service_account_auth_file = config["trailblazer"]["service_account_auth_file"]
        self.host = config["trailblazer"]["host"]

    @property
    def auth_header(self) -> dict:
        signer = RSASigner.from_service_account_file(self.service_account_auth_file)
        payload = {"email": self.service_account}
        jwt_token = jwt.encode(signer=signer, payload=payload).decode("ascii")
        return {"Authorization": f"Bearer {jwt_token}"}

    def query_trailblazer(
        self, command: str, request_body: dict, method: str = APIMethods.POST
    ) -> Any:
        url = self.host + "/" + command
        LOG.debug(f"REQUEST HEADER {self.auth_header}")
        LOG.debug(f"{method}: URL={url}; JSON={request_body}")

        response = APIRequest.api_request_from_content(
            api_method=method, url=url, headers=self.auth_header, json=request_body
        )

        LOG.debug(f"RESPONSE STATUS CODE {response.status_code}")
        if not response.ok:
            raise TrailblazerAPIHTTPError(
                f"Request {command} failed with status code {response.status_code}: {response.text}"
            )
        LOG.debug(f"RESPONSE BODY {response.text}")
        return ReadStream.get_content_from_stream(file_format=FileFormat.JSON, stream=response.text)

    def analyses(
        self,
        case_id: str = None,
        query: str = None,
        status: str = None,
        deleted: bool = None,
        temp: bool = False,
        before: dt.datetime = None,
        is_visible: bool = None,
        family: str = None,
        data_analysis: Pipeline = None,
    ) -> list:
        request_body = {
            "case_id": case_id,
            "status": status,
            "query": query,
            "deleted": deleted,
            "temp": temp,
            "before": str(before) if before else None,
            "is_visible": is_visible,
            "family": family,
            "data_analysis": data_analysis.upper() if data_analysis else None,
        }
        response = self.query_trailblazer(command="query-analyses", request_body=request_body)
        if response:
            if isinstance(response, list):
                return [TrailblazerAnalysis.parse_obj(analysis) for analysis in response]
            if isinstance(response, dict):
                return [TrailblazerAnalysis.parse_obj(response)]
        return response

    def get_latest_analysis(self, case_id: str) -> Optional[TrailblazerAnalysis]:
        request_body = {
            "case_id": case_id,
        }
        response = self.query_trailblazer(command="get-latest-analysis", request_body=request_body)
        if response:
            return TrailblazerAnalysis.parse_obj(response)

    def find_analysis(
        self, case_id: str, started_at: dt.datetime, status: str
    ) -> Optional[TrailblazerAnalysis]:
        request_body = {"case_id": case_id, "started_at": str(started_at), "status": status}
        response = self.query_trailblazer(command="find-analysis", request_body=request_body)
        if response:
            return TrailblazerAnalysis.parse_obj(response)

    def get_latest_analysis_status(self, case_id: str) -> Optional[str]:
        latest_analysis = self.get_latest_analysis(case_id=case_id)
        if latest_analysis:
            return latest_analysis.status

    def has_latest_analysis_started(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) in self.__STARTED_STATUSES

    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) in self.__ONGOING_STATUSES

    def is_latest_analysis_completed(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) == AnalysisStatus.COMPLETED

    def is_latest_analysis_qc(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) == AnalysisStatus.QC

    def delete_analysis(self, analysis_id: str, force: bool = False) -> None:
        """Raises TrailblazerAPIHTTPError"""
        request_body = {"analysis_id": analysis_id, "force": force}
        self.query_trailblazer(command="delete-analysis", request_body=request_body)

    def mark_analyses_deleted(self, case_id: str) -> Optional[list]:
        """Mark all analyses for case deleted without removing analysis files"""
        request_body = {
            "case_id": case_id,
        }
        response = self.query_trailblazer(
            command="mark-analyses-deleted", request_body=request_body
        )
        if response:
            if isinstance(response, list):
                return [TrailblazerAnalysis.parse_obj(analysis) for analysis in response]
            if isinstance(response, dict):
                return [TrailblazerAnalysis.parse_obj(response)]

    def add_pending_analysis(
        self,
        case_id: str,
        analysis_type: str,
        config_path: str,
        out_dir: str,
        slurm_quality_of_service: SlurmQos,
        email: str = None,
        data_analysis: Pipeline = None,
        ticket: str = None,
        workflow_manager: str = WorkflowManager.Slurm,
    ) -> TrailblazerAnalysis:
        request_body = {
            "case_id": case_id,
            "email": email,
            "type": analysis_type,
            "config_path": config_path,
            "out_dir": out_dir,
            "priority": slurm_quality_of_service,
            "data_analysis": str(data_analysis).upper(),
            "ticket": ticket,
            "workflow_manager": workflow_manager,
        }
        LOG.debug("Submitting job to Trailblazer: %s", request_body)
        response = self.query_trailblazer(command="add-pending-analysis", request_body=request_body)
        if response:
            return TrailblazerAnalysis.parse_obj(response)

    def set_analysis_uploaded(self, case_id: str, uploaded_at: datetime) -> None:
        """Set a uploaded at date for a trailblazer analysis."""
        request_body = {"case_id": case_id, "uploaded_at": str(uploaded_at)}

        LOG.debug(f"Setting analysis uploaded at for {request_body}")
        LOG.info(f"{case_id} - uploaded at set to {uploaded_at}")
        self.query_trailblazer(
            command="set-analysis-uploaded", request_body=request_body, method=APIMethods.PUT
        )

    def set_analysis_status(self, case_id: str, status: str) -> datetime:
        """Set an analysis to a given status."""
        request_body = {"case_id": case_id, "status": status}

        LOG.debug(f"Request body: {request_body}")
        LOG.info(f"Setting analysis status to {status} for case {case_id}")
        self.query_trailblazer(
            command="set-analysis-status", request_body=request_body, method=APIMethods.PUT
        )

    def add_comment(self, case_id: str, comment: str):
        """Adding a comment to a trailblazer analysis."""
        request_body = {"case_id": case_id, "comment": comment}

        LOG.debug(f"Request body: {request_body}")
        LOG.info(f"Adding comment: {comment} to case {case_id}")
        self.query_trailblazer(
            command="add-comment", request_body=request_body, method=APIMethods.PUT
        )



import json

from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import Bcl2FastqSequencingMetrics
from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq import parse_bcl2fastq_sequencing_metrics


def test_parse_valid_bcl2fastq_sequencing_metrics(tmp_path, valid_bcl2fastq_metrics_data):
    """Test that a valid stats.json file as generated by bcl2fastq can be parsed correctly."""
    # GIVEN a valid stats.json file
    stats_json_path = tmp_path / "stats.json"
    stats_json_path.write_text(json.dumps(valid_bcl2fastq_metrics_data))

    # WHEN parsing the valid stats.json file
    result = parse_bcl2fastq_sequencing_metrics(str(stats_json_path))

    # THEN a Bcl2FastqSequencingMetrics model is returned
    assert isinstance(result, Bcl2FastqSequencingMetrics)

    # THEN the Bcl2FastqSequencingMetrics model contains the expected data
    assert result.flowcell == valid_bcl2fastq_metrics_data["Flowcell"]
    assert result.run_number == valid_bcl2fastq_metrics_data["RunNumber"]
    assert result.run_id == valid_bcl2fastq_metrics_data["RunId"]



import logging

from typing import List, Optional

from cg.apps.cgstats.db.models import Flowcell
from cg.apps.cgstats.stats import StatsAPI

log = logging.getLogger(__name__)


def delete_flowcell(manager: StatsAPI, flowcell_name: str):
    flow_cell: Optional[Flowcell] = manager.find_handler.get_flow_cell_by_name(
        flow_cell_name=flowcell_name
    )

    if flow_cell:
        log.info("Removing entry %s in from cgstats", flow_cell.flowcellname)
        manager.delete(flow_cell)
        manager.commit()



from pathlib import Path

from click.testing import CliRunner, Result

from cg.apps.demultiplex.sample_sheet.validate import get_sample_sheet_from_file
from cg.cli.demultiplex.sample_sheet import validate_sample_sheet

from cg.constants import EXIT_SUCCESS, FileExtensions
from cg.constants.demultiplexing import BclConverter


def test_validate_non_existing_sample_sheet(cli_runner: CliRunner, sample_sheet_context: dict):
    """Test validate sample sheet when sample sheet does not exist."""

    # GIVEN a cli runner
    # GIVEN a sample sheet file that does not exist
    sample_sheet: Path = Path("a_sample_sheet_that_does_not_exist.csv")
    assert sample_sheet.exists() is False

    # WHEN validating the sample sheet
    result = cli_runner.invoke(validate_sample_sheet, [str(sample_sheet)], obj=sample_sheet_context)

    # THEN assert that it exits with a non-zero exit code
    assert result.exit_code != EXIT_SUCCESS
    # THEN assert the correct information was communicated
    assert f'Path "{sample_sheet.name}" does not exist' in result.output


def test_validate_sample_sheet_wrong_file_type(
    cli_runner: CliRunner, sample_sheet_context: dict, novaseq_run_parameters: Path, caplog
):
    """Test validate sample sheet when sample sheet is in wrong format."""
    # GIVEN a existing file in the wrong file format
    sample_sheet: Path = novaseq_run_parameters
    assert sample_sheet.exists()
    assert sample_sheet.suffix != FileExtensions.CSV

    # WHEN validating the sample sheet
    result: Result = cli_runner.invoke(
        validate_sample_sheet, [str(sample_sheet)], obj=sample_sheet_context
    )

    # THEN assert it exits with a non-zero exit code
    assert result.exit_code != EXIT_SUCCESS

    # THEN assert the correct information was communicated
    assert "seems to be in wrong format" in caplog.text


def test_validate_correct_bcl2fastq_sample_sheet(
    cli_runner: CliRunner, sample_sheet_context: dict, novaseq_bcl2fastq_sample_sheet_path: Path
):
    """Test validate sample sheet when using a bcl2fastq sample sheet."""

    # GIVEN the path to a bcl2fastq sample sheet that exists
    sample_sheet: Path = novaseq_bcl2fastq_sample_sheet_path
    assert sample_sheet.exists()

    # GIVEN that the sample sheet is correct
    get_sample_sheet_from_file(
        infile=sample_sheet,
        bcl_converter=BclConverter.BCL2FASTQ,
    )

    # WHEN validating the sample sheet
    result: Result = cli_runner.invoke(
        validate_sample_sheet, [str(sample_sheet)], obj=sample_sheet_context
    )

    # THEN assert that it exits successfully
    assert result.exit_code == EXIT_SUCCESS


def test_validate_correct_dragen_sample_sheet(
    cli_runner: CliRunner, sample_sheet_context: dict, novaseq_dragen_sample_sheet_path: Path
):
    """Test validate sample sheet when using a Dragen sample sheet."""

    # GIVEN the path to a bcl2fastq sample sheet that exists
    sample_sheet: Path = novaseq_dragen_sample_sheet_path
    assert sample_sheet.exists()

    # GIVEN that the sample sheet is correct
    get_sample_sheet_from_file(infile=sample_sheet, bcl_converter=BclConverter.DRAGEN)

    # WHEN validating the sample sheet
    result: Result = cli_runner.invoke(
        validate_sample_sheet,
        [str(sample_sheet), "-b", BclConverter.DRAGEN.value],
        obj=sample_sheet_context,
    )

    # THEN assert that it exits successfully
    assert result.exit_code == EXIT_SUCCESS



from typing import List

from cg.apps.demultiplex.sample_sheet import dummy_sample, index
from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.lims.samplesheet import LimsFlowcellSample


def test_get_valid_indexes():
    # GIVEN a sample sheet api

    # WHEN fetching the indexes
    indexes: List[Index] = index.get_valid_indexes()

    # THEN assert that the indexes are correct
    assert len(indexes) > 0
    assert isinstance(indexes[0], Index)


def test_get_dummy_sample_name():
    # GIVEN a raw sample name from the index file
    raw_sample_name = "D10 - D710-D504 (TCCGCGAA-GGCTCTGA)"

    # WHEN converting it to a dummy sample name
    dummy_sample_name: str = dummy_sample.dummy_sample_name(raw_sample_name)

    # THEN assert the correct name was created
    assert dummy_sample_name == "D10---D710-D504--TCCGCGAA-GGCTCTGA-"


def test_get_dummy_sample(bcl2fastq_flow_cell_id: str, index_obj: Index):
    # GIVEN some dummy sample data

    # WHEN creating the dummy sample for a bcl2fastq samplesheet
    dummy_sample_obj: LimsFlowcellSample = dummy_sample.dummy_sample(
        flowcell=bcl2fastq_flow_cell_id,
        dummy_index=index_obj.sequence,
        lane=1,
        name=index_obj.name,
        bcl_converter="bcl2fastq",
    )

    # THEN assert the sample id was correct
    assert dummy_sample_obj.sample_id == dummy_sample.dummy_sample_name(index_obj.name)



from pathlib import Path
from typing import List

from click import testing

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.lims.samplesheet import (
    LimsFlowcellSampleBcl2Fastq,
    LimsFlowcellSampleDragen,
)
from cg.cli.demultiplex.sample_sheet import create_sheet
from cg.constants.demultiplexing import BclConverter
from cg.constants.process import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCell

FLOW_CELL_FUNCTION_NAME: str = "cg.cli.demultiplex.sample_sheet.flowcell_samples"


def test_create_sample_sheet_no_run_parameters(
    cli_runner: testing.CliRunner,
    flow_cell_working_directory_no_run_parameters: Path,
    sample_sheet_context: CGConfig,
    lims_novaseq_bcl2fastq_samples: List[LimsFlowcellSampleBcl2Fastq],
    caplog,
    mocker,
):
    # GIVEN a folder with a non-existing sample sheet
    flowcell_object: FlowCell = FlowCell(flow_cell_working_directory_no_run_parameters)
    assert flowcell_object.run_parameters_path.exists() is False

    # GIVEN flow cell samples
    mocker.patch(
        FLOW_CELL_FUNCTION_NAME,
        return_value=lims_novaseq_bcl2fastq_samples,
    )

    # GIVEN a demux API context
    demux_api: DemultiplexingAPI = sample_sheet_context.demultiplex_api
    demux_api.run_dir: Path = flow_cell_working_directory_no_run_parameters.parent
    sample_sheet_context.demultiplex_api_: DemultiplexingAPI = demux_api

    # WHEN running the create sample sheet command
    result: testing.Result = cli_runner.invoke(
        create_sheet, [flowcell_object.full_name], obj=sample_sheet_context
    )

    # THEN the process exits with a non-zero exit code
    assert result.exit_code != EXIT_SUCCESS

    # THEN the correct information is communicated
    assert "Could not find run parameters file" in caplog.text


def test_create_bcl2fastq_sample_sheet(
    cli_runner: testing.CliRunner,
    flow_cell_working_directory: Path,
    sample_sheet_context: CGConfig,
    lims_novaseq_bcl2fastq_samples: List[LimsFlowcellSampleBcl2Fastq],
    mocker,
):
    # GIVEN a flowcell directory with some run parameters
    flowcell: FlowCell = FlowCell(flow_cell_working_directory)
    assert flowcell.run_parameters_path.exists()

    # GIVEN that there is no sample sheet present
    assert not flowcell.sample_sheet_exists()

    # GIVEN flow cell samples
    mocker.patch(
        FLOW_CELL_FUNCTION_NAME,
        return_value=lims_novaseq_bcl2fastq_samples,
    )
    # GIVEN a lims api that returns some samples

    # WHEN creating a sample sheet
    result = cli_runner.invoke(
        create_sheet, [str(flow_cell_working_directory)], obj=sample_sheet_context
    )

    # THEN the process finishes successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN the sample sheet was created
    assert flowcell.sample_sheet_exists()

    # THEN the sample sheet is on the correct format
    assert flowcell.validate_sample_sheet()


def test_create_dragen_sample_sheet(
    cli_runner: testing.CliRunner,
    flow_cell_working_directory: Path,
    sample_sheet_context: CGConfig,
    lims_novaseq_dragen_samples: List[LimsFlowcellSampleDragen],
    mocker,
):
    # GIVEN a flowcell directory with some run parameters
    flowcell: FlowCell = FlowCell(flow_cell_working_directory, bcl_converter=BclConverter.DRAGEN)
    assert flowcell.run_parameters_path.exists()

    # GIVEN that there is no sample sheet present
    assert not flowcell.sample_sheet_exists()

    # GIVEN flow cell samples
    mocker.patch(
        FLOW_CELL_FUNCTION_NAME,
        return_value=lims_novaseq_dragen_samples,
    )
    # GIVEN a lims api that returns some samples

    # WHEN creating a sample sheet
    result = cli_runner.invoke(
        create_sheet,
        [str(flow_cell_working_directory), "-b", BclConverter.DRAGEN],
        obj=sample_sheet_context,
    )

    # THEN the process finishes successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN the sample sheet was created
    assert flowcell.sample_sheet_exists()

    # THEN the sample sheet is on the correct format
    assert flowcell.validate_sample_sheet()



"""Constants specific for compression."""
import datetime
from typing import List

# Constants for crunchy
FASTQ_FIRST_READ_SUFFIX: str = "_R1_001.fastq.gz"
FASTQ_SECOND_READ_SUFFIX: str = "_R2_001.fastq.gz"
FLAG_PATH_SUFFIX: str = ".crunchy.txt"
MAX_READS_PER_GB: int = 18_000_000
CRUNCHY_MIN_GB_PER_PROCESS: int = 30
PENDING_PATH_SUFFIX: str = ".crunchy.pending.txt"


# Number of days until FASTQs counts as old
FASTQ_DELTA = 21
FASTQ_DATETIME_DELTA = datetime.timedelta(days=FASTQ_DELTA)

PROBLEMATIC_CASES = [
    "causalmite",
    "deepcub",
    "expertalien",
    "fluenteagle",
    "grandkoi",
    "lovingmayfly",
    "loyalegret",
    "modernbee",
    "proudcollie",
    "richalien",
    "suremako",
    "wisestork",
]

# List of cases used for validation that we should skip
BALSAMIC_VALIDATION_CASES = [
    "setamoeba",  # BALSAMIC validation case tumor-only panel
    "sweetelf",  # BALSAMIC positive control tumor-only panel
    "poeticghoul",  # BALSAMIC positive control tumor-only panel
    "equalbug",  # UMI seracare validation case (AF 0.5%) tumor-normal panel
    "stableraven",  # UMI seracare validation case (AF 1%) tumor-normal panel
    "uphippo",  # UMI seracare validation case (AF 0.5%) tumor-only panel
    "cleanfowl",  # BALSAMIC validation case, HD829 reference for FLT3 Ascertation
    "proudsquid",  # BALSAMIC validation case, HD829 reference for FLT3 Ascertation
    "modestjaguar",  # BALSAMIC validation case, HD829 reference for FLT3 Ascertation
    "dearmarmot",  # BALSAMIC validation case, HD829 reference for FLT3 Ascertation
    "holykid",  # BALSAMIC validation case, HD829 reference for FLT3 Ascertation
    "civilsole",  # BALSAMIC validation case tumor-only wgs
    "fleetjay",  # BALSAMIC validation case tumor-normal wgs
    "grandmarmot",  # BALSAMIC validation case tumor-normal wgs
    "unitedbeagle",  # BALSAMIC validation case tumor-normal panel
    "rightthrush",  # BALSAMIC validation case tumor-only wes
    "properpigeon",  # BALSAMIC validation case tumor-normal wes
    "eagerox",  # BALSAMIC validation case from cust087, tumor-only panel
    "casualweasel",  # BALSAMIC validation case from cust087, tumor-only panel
    "acetuna",  # BALSAMIC validation case from cust087, tumor-only panel
    "suitedsnake",  # BALSAMIC validation case from cust087, tumor-only panel
    "savinghorse",  # BALSAMIC validation case from cust087, tumor-only panel
    "rightpup",  # BALSAMIC validation case from cust087, tumor-only panel
    "sureroughy",  # BALSAMIC validation case from cust087, tumor-only panel
    "notedshark",  # BALSAMIC validation case from cust127, tumor-normal WGS (SV inversion positive control)
    "wholewhale",  # BALSAMIC validation case from cust143, tumor-normal WGS (SV translocation positive control)
    "largeturtle",  # BALSAMIC validation case from cust143, tumor-normal WGS (SV translocation positive control)
    "lightkodiak",  # BALSAMIC validation case from cust143, tumor-normal WGS (SV inversion positive control)
    "wholecivet",  # BALSAMIC validation case from cust110, tumor-normal WGS (SV inversion positive control)
    "upwardstork",  # BALSAMIC validation case from cust110, tumor-normal WGS (SV deletion positive control)
    "suitedgrub",  # BALSAMIC validation case from cust110, tumor-normal WGS (SV deletion positive control)
]

FLUFFY_VALIDATION_CASES = [
    "simplesalmon",  # Chromosome 13, 18, 21 Suspected
]

MIP_VALIDATION_CASES = [
    "brightcaiman",  # DNA rare disease positive control
    "casualgannet",  # DNA rare disease positive control
    "civilkoala",  # RNA rare disease positive control
    "cleanshrimp",  # DNA rare disease positive control
    "drivenmolly",  # RNA rare disease positive control
    "easybeetle",  # DNA rare disease positive control
    "epicasp",  # DNA rare disease positive control
    "expertmole",  # RNA rare disease positive control
    "finequagga",  # RNA rare disease positive control
    "firstfawn",  # DNA rare disease positive control
    "gladthrush",  # DNA rare disease positive control
    "helpedfilly",  # DNA rare disease positive control
    "hotskink",  # DNA rare disease positive control
    "inferret",  # DNA rare disease positive control
    "intentcorgi",  # DNA rare disease positive control
    "intentmayfly",  # DNA rare disease positive control
    "justhusky",  # DNA rare disease positive control
    "kindcaiman",  # DNA rare disease positive control
    "lightprawn",  # DNA rare disease positive control
    "livingox",  # DNA rare disease positive control
    "newaphid",  # RNA rare disease positive control
    "nextjackal",  # DNA rare disease positive control
    "modernmule",  # DNA rare disease positive control
    "moralcattle",  # RNA rare disease positive control
    "onemite",  # DNA rare disease positive control
    "opencow",  # DNA rare disease positive control
    "proudcougar",  # DNA rare disease positive control
    "rightmacaw",  # DNA rare disease positive control
    "safeguinea",  # DNA rare disease positive control
    "sharpparrot",  # RNA rare disease positive control
    "sharppigeon",  # DNA rare disease positive control
    "sharpwhale",  # DNA rare disease positive control
    "stillant",  # DNA rare disease positive control
    "smoothboa",  # RNA rare disease positive control
    "strongbison",  # DNA rare disease positive control
    "tenderoriole",  # DNA rare disease positive control
    "topsrhino",  # DNA rare disease positive control
    "usablemarten",  # DNA rare disease positive control
    "vitalmouse",  # DNA rare disease positive control
]

# List of cases used for validation that we should skip
RNAFUSION_VALIDATION_CASES = [
    "cuddlyhen",  # downsampled seracare commercial sample
    "oncrab",
    "daringowl",
    "growndoe",
    "stablemoray",
    "truemole",
    "rightmoray",
    "handyturtle",
    "inlab",
    "honestswine",
    "holyrodent",
    "expertboar",
    "bossmink",
    "ampleray",
    "nearbyjoey",
]

TAXPROFILER_VALIDATION_CASES: List[str] = [
    "richurchin",
]

OTHER_VALIDATION_CASES = [
    "bigdrum",
    "busycolt",
    "daringpony",
    "frankhusky",
    "gamedeer",
    "hotviper",
    "keencalf",
    "keenviper",
    "luckyhog",
    "maturejay",  # sars-cov-2 case
    "meetpossum",
    "mintbaboon",
    "mintyeti",
    "proeagle",
    "propercoral",
    "pumpedcat",
    "strongman",
    "truecoyote",
]

CASES_TO_IGNORE = (
    PROBLEMATIC_CASES
    + OTHER_VALIDATION_CASES
    + BALSAMIC_VALIDATION_CASES
    + FLUFFY_VALIDATION_CASES
    + MIP_VALIDATION_CASES
    + RNAFUSION_VALIDATION_CASES
    + TAXPROFILER_VALIDATION_CASES
)



"""Class to hold information about scout load config"""

from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, validator
from typing_extensions import Literal


class ChromographImages(BaseModel):
    autozygous: Optional[str] = None
    coverage: Optional[str] = None
    upd_regions: Optional[str] = None
    upd_sites: Optional[str] = None


class Reviewer(BaseModel):
    alignment: Optional[str] = None
    alignment_index: Optional[str] = None
    vcf: Optional[str] = None
    catalog: Optional[str] = None


class ScoutIndividual(BaseModel):
    alignment_path: Optional[str] = None
    analysis_type: Literal[
        "external",
        "mixed",
        "panel",
        "panel-umi",
        "unknown",
        "wes",
        "wgs",
        "wts",
    ] = None
    capture_kit: Optional[str] = None
    confirmed_parent: Optional[bool] = None
    confirmed_sex: Optional[bool] = None
    father: Optional[str] = None
    mother: Optional[str] = None
    phenotype: Optional[str] = None
    sample_id: str = None
    sample_name: Optional[str] = None
    sex: Optional[str] = None
    subject_id: Optional[str] = None
    tissue_type: Optional[str] = None

    @validator("sample_id", "sex", "analysis_type")
    def field_not_none(cls, value):
        if value is None:
            raise ValueError("sample_id, sex and analysis_type can not be None")
        return value

    class Config:
        validate_assignment = True


class ScoutMipIndividual(ScoutIndividual):
    mt_bam: Optional[str] = None
    chromograph_images: ChromographImages = ChromographImages()
    reviewer: Reviewer = Reviewer()
    rhocall_bed: Optional[str] = None
    rhocall_wig: Optional[str] = None
    tiddit_coverage_wig: Optional[str] = None
    upd_regions_bed: Optional[str] = None
    upd_sites_bed: Optional[str] = None
    vcf2cytosure: Optional[str] = None
    mitodel_file: Optional[str] = None


class ScoutCancerIndividual(ScoutIndividual):
    tumor_type: Optional[str] = None
    tmb: Optional[str] = None
    msi: Optional[str] = None
    tumor_purity: float = 0
    vcf2cytosure: Optional[str] = None


class ScoutLoadConfig(BaseModel):
    owner: str = None
    family: str = None
    family_name: Optional[str] = None
    synopsis: Optional[str] = None
    phenotype_terms: Optional[List[str]] = None
    phenotype_groups: Optional[List[str]] = None
    gene_panels: Optional[List[str]] = None
    default_gene_panels: List[str] = []
    cohorts: Optional[List[str]] = None
    human_genome_build: str = None

    rank_model_version: Optional[str] = None
    rank_score_threshold: int = None
    sv_rank_model_version: Optional[str] = None
    analysis_date: datetime = None
    samples: List[ScoutIndividual] = []

    delivery_report: Optional[str] = None
    coverage_qc_report: Optional[str] = None
    cnv_report: Optional[str] = None
    multiqc: Optional[str] = None
    track: Literal["rare", "cancer"] = "rare"

    @validator("owner", "family")
    def field_not_none(cls, value):
        if value is None:
            raise ValueError("Owner and family can not be None")
        return value

    class Config:
        validate_assignment = True


class BalsamicLoadConfig(ScoutLoadConfig):
    madeline: Optional[str]
    vcf_cancer: str = None
    vcf_cancer_sv: Optional[str] = None
    vcf_cancer_research: Optional[str] = None
    vcf_cancer_sv_research: Optional[str] = None
    samples: List[ScoutCancerIndividual] = []

    @validator("vcf_cancer")
    def check_mandatory_files(cls, vcf):
        if vcf is None:
            raise ValueError("Vcf can not be none")
        return vcf


class BalsamicUmiLoadConfig(BalsamicLoadConfig):
    pass


class MipLoadConfig(ScoutLoadConfig):
    chromograph_image_files: Optional[List[str]]
    chromograph_prefixes: Optional[List[str]]
    madeline: Optional[str] = None
    peddy_check: Optional[str] = None
    peddy_ped: Optional[str] = None
    peddy_sex: Optional[str] = None
    samples: List[ScoutMipIndividual] = []
    smn_tsv: Optional[str] = None
    variant_catalog: Optional[str] = None
    vcf_mei: Optional[str] = None
    vcf_mei_research: Optional[str] = None
    vcf_snv: str = None
    vcf_snv_research: Optional[str] = None
    vcf_str: Optional[str] = None
    vcf_sv: Optional[str] = None
    vcf_sv_research: Optional[str] = None

    @validator("vcf_snv", "vcf_sv", "vcf_snv_research", "vcf_sv_research")
    def check_mandatory_files(cls, vcf):
        if vcf is None:
            raise ValueError("Mandatory vcf can not be None")
        return vcf


class RnafusionLoadConfig(ScoutLoadConfig):
    multiqc_rna: Optional[str] = None
    gene_fusion: Optional[str] = None
    gene_fusion_report_research: Optional[str] = None
    RNAfusion_inspector: Optional[str] = None
    RNAfusion_inspector_research: Optional[str] = None
    RNAfusion_report: Optional[str] = None
    RNAfusion_report_research: Optional[str] = None
    samples: List[ScoutCancerIndividual] = []



from datetime import datetime

import click
import pytest
from cgmodels.cg.constants import Pipeline

from cg.cli.generate.report.base import generate_delivery_report
from cg.models.cg_config import CGConfig
from tests.mocks.report import MockMipDNAReportAPI, MockMipDNAAnalysisAPI


@pytest.fixture(name="mip_dna_context")
def fixture_mip_dna_context(cg_context, helpers, case_id, real_housekeeper_api) -> CGConfig:
    """MIP DNA context fixture"""

    store = cg_context.status_db
    cg_context.housekeeper_api_ = real_housekeeper_api
    cg_context.meta_apis["report_api"] = MockMipDNAReportAPI(
        cg_context, MockMipDNAAnalysisAPI(config=cg_context)
    )

    # Add app tag, case, sample, analysis and relationships to DB
    helpers.ensure_application_version(
        store=store,
        application_tag="WGSA",
        prep_category="wgs",
    )
    case = helpers.add_case(
        store=store,
        data_analysis=Pipeline.MIP_DNA,
        internal_id=case_id,
    )
    sample = helpers.add_sample(
        store=store,
        customer_id="cust000",
        application_tag="WGSA",
        application_type="wgs",
    )
    helpers.add_analysis(
        store=store,
        case=case,
        pipeline=Pipeline.MIP_DNA,
        delivery_reported_at=datetime.now(),
        started_at=datetime.now(),
    )
    helpers.add_relationship(
        store=store,
        sample=sample,
        case=case,
        status="affected",
    )

    return cg_context


@pytest.fixture(name="delivery_report_click_context")
def fixture_delivery_report_click_context(mip_dna_context) -> click.Context:
    """Click delivery report context fixture"""

    return click.Context(generate_delivery_report, obj=mip_dna_context)



"""Module for Taxprofiler Analysis API."""

import logging
from cg.constants import Pipeline
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class TaxprofilerAnalysisAPI(AnalysisAPI):
    """Handles communication between Taxprofiler processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.TAXPROFILER,
    ):
        super().__init__(config=config, pipeline=pipeline)



from pathlib import Path
from typing import List

import pytest


@pytest.fixture(name="input_file")
def fixture_input_file_path() -> Path:
    """input file"""
    return Path("/path/to/input.file")


@pytest.fixture(name="output_file")
def fixture_output_file_path() -> Path:
    """output file"""
    return Path("/path/to/output.file")


@pytest.fixture(name="temporary_passphrase")
def fixture_temporary_passphrase() -> str:
    """temporary passphrase file"""
    return "tmp/tmp_test_passphrase"


@pytest.fixture(name="test_command")
def fixture_test_command() -> List:
    """Return a CLI command in the list format required by the Process API"""
    return ["test", "command"]


@pytest.fixture(name="encryption_key_file")
def fixture_encryption_key_file() -> Path:
    """Return an encryption key file Path object"""
    return Path("/path/to/file.key")


@pytest.fixture(name="encrypted_key_file")
def fixture_encrypted_key_file() -> Path:
    """Return an encrypted encryption key file Path object"""
    return Path("/path/to/encryption.key.gpg")


@pytest.fixture(name="spring_file_path")
def fixture_spring_file_path() -> Path:
    """Return an spring file Path object"""
    return Path("/path/to/file.spring")


@pytest.fixture(name="checksum_spring_file_path")
def fixture_checksum_spring_file_path() -> Path:
    """Return an spring file Path object"""
    return Path("/path/to/file.spring.tmp")


@pytest.fixture(name="encrypted_spring_file_path")
def fixture_encrypted_spring_file_path() -> Path:
    """Return an encrypted spring file Path object"""
    return Path("/path/to/file.spring.gpg")


@pytest.fixture(name="checksum_result_1")
def fixture_checksum_result_1() -> str:
    """fixture that can be used as a checksum result"""
    return "a1b2c3"


@pytest.fixture(name="checksum_result_2")
def fixture_checksum_result_2() -> str:
    """fixture that can be used as a checksum result"""
    return "x7y8z9"


@pytest.fixture(name="encryption_user_id")
def fixture_encryption_user_id() -> str:
    """Encryption user ID fixture"""
    return "Clinical Genomics"


@pytest.fixture(name="cipher_algorithm")
def fixture_cipher_algorithm() -> str:
    """Cipher algorithm fixture"""
    return "AES256"


@pytest.fixture(name="asymmetric_encryption_command")
def fixture_asymmetric_encryption_command(encryption_user_id, output_file, input_file) -> List:
    """Asymmetric encryption_command fixture"""
    return [
        "--encrypt",
        "--recipient",
        encryption_user_id,
        "-o",
        str(output_file),
        str(input_file),
    ]


@pytest.fixture(name="asymmetric_decryption_command")
def fixture_asymmetric_decryption_command(
    cipher_algorithm, encryption_user_id, output_file, input_file
) -> List:
    """Asymmetric decryption_command fixture"""
    return [
        "--decrypt",
        "--batch",
        "--cipher-algo",
        cipher_algorithm,
        "--passphrase",
        encryption_user_id,
        "-o",
        str(output_file),
        str(input_file),
    ]


@pytest.fixture(name="symmetric_encryption_command")
def fixture_symmetric_encryption_command(
    temporary_passphrase, input_file, output_file, spring_file_path, cipher_algorithm
) -> List:
    """Symmetric encryption_command fixture"""
    return [
        "--symmetric",
        "--cipher-algo",
        cipher_algorithm,
        "--batch",
        "--compress-algo",
        "None",
        "--passphrase-file",
        temporary_passphrase,
        "-o",
        str(output_file),
        str(input_file),
    ]


@pytest.fixture(name="symmetric_decryption_command")
def fixture_symmetric_decryption_command(
    cipher_algorithm,
    encryption_key_file,
    input_file,
    output_file,
) -> List:
    """Symmetric decryption_command fixture"""
    return [
        "--decrypt",
        "--cipher-algo",
        cipher_algorithm,
        "--batch",
        "--passphrase-file",
        str(encryption_key_file),
        "-o",
        str(output_file),
        str(input_file),
    ]


@pytest.fixture(name="spring_symmetric_encryption_command")
def fixture_spring_symmetric_encryption_command(
    temporary_passphrase, encrypted_spring_file_path, spring_file_path, cipher_algorithm
) -> List:
    """Symmetric encryption_command fixture"""
    return [
        "--symmetric",
        "--cipher-algo",
        cipher_algorithm,
        "--batch",
        "--compress-algo",
        "None",
        "--passphrase-file",
        temporary_passphrase,
        "-o",
        str(encrypted_spring_file_path),
        str(spring_file_path),
    ]


@pytest.fixture(name="key_asymmetric_encryption_command")
def fixture_key_asymmetric_encryption_command(
    encryption_user_id, encrypted_key_file, temporary_passphrase
) -> List:
    """Asymmetric encryption_command fixture"""
    return [
        "--encrypt",
        "--recipient",
        encryption_user_id,
        "-o",
        str(encrypted_key_file),
        str(temporary_passphrase),
    ]


@pytest.fixture(name="spring_symmetric_decryption_command")
def fixture_spring_symmetric_decryption_command(
    cipher_algorithm,
    encryption_key_file,
    spring_file_path,
    encrypted_spring_file_path,
) -> List:
    """Symmetric decryption_command fixture"""
    return [
        "--decrypt",
        "--cipher-algo",
        cipher_algorithm,
        "--batch",
        "--passphrase-file",
        str(encryption_key_file),
        "-o",
        str(spring_file_path),
        str(encrypted_spring_file_path),
    ]


@pytest.fixture(name="key_asymmetric_decryption_command")
def fixture_key_asymmetric_decryption_command(
    cipher_algorithm, encryption_user_id, encrypted_key_file, encryption_key_file
) -> List:
    """Asymmetric encryption_command fixture"""
    return [
        "--decrypt",
        "--batch",
        "--cipher-algo",
        cipher_algorithm,
        "--passphrase",
        encryption_user_id,
        "-o",
        str(encryption_key_file),
        str(encrypted_key_file),
    ]



"""This module contains tests for the BCLConvert metrics parser."""
from pathlib import Path

from cg.apps.sequencing_metrics_parser.parsers.bcl_convert import BclConvertMetricsParser
from cg.apps.sequencing_metrics_parser.models.bcl_convert import (
    BclConvertQualityMetrics,
    BclConvertDemuxMetrics,
    BclConvertSampleSheetData,
    BclConvertAdapterMetrics,
    BclConvertRunInfo,
)


def test_parse_bcl_convert_metrics(
    bcl_convert_quality_metric_file_path,
    bcl_convert_demux_metric_file_path,
    bcl_convert_sample_sheet_file_path,
    bcl_convert_adapter_metrics_file_path,
    bcl_convert_run_info_file_path,
):
    """Test to parse BCLConvert metrics."""
    # GIVEN paths to a BCLConvert metrics files
    # WHEN parsing the files
    bcl_convert_metrics_parser = BclConvertMetricsParser(
        bcl_convert_quality_metrics_file_path=bcl_convert_quality_metric_file_path,
        bcl_convert_demux_metrics_file_path=bcl_convert_demux_metric_file_path,
        bcl_convert_sample_sheet_file_path=bcl_convert_sample_sheet_file_path,
        bcl_convert_adapter_metrics_file_path=bcl_convert_adapter_metrics_file_path,
        bcl_convert_run_info_file_path=bcl_convert_run_info_file_path,
    )

    # THEN assert that the metrics are parsed
    assert bcl_convert_metrics_parser.quality_metrics
    assert isinstance(bcl_convert_metrics_parser.quality_metrics[0], BclConvertQualityMetrics)
    assert bcl_convert_metrics_parser.demux_metrics
    assert isinstance(bcl_convert_metrics_parser.demux_metrics[0], BclConvertDemuxMetrics)
    assert bcl_convert_metrics_parser.sample_sheet
    assert isinstance(bcl_convert_metrics_parser.sample_sheet[0], BclConvertSampleSheetData)
    assert bcl_convert_metrics_parser.adapter_metrics
    assert isinstance(bcl_convert_metrics_parser.adapter_metrics[0], BclConvertAdapterMetrics)
    assert bcl_convert_metrics_parser.run_info
    assert isinstance(bcl_convert_metrics_parser.run_info, BclConvertRunInfo)


def test_parse_bcl_convert_quality_metrics(
    parsed_bcl_convert_metrics: BclConvertMetricsParser,
    bcl_convert_quality_metric_model_with_data: BclConvertQualityMetrics,
):
    """Test to parse BCLConvert quality metrics."""
    # GIVEN a parsed BCLConvert metrics

    # ASSERT that the parsed quality metrics are correct
    quality_metrics_model: BclConvertQualityMetrics = parsed_bcl_convert_metrics.quality_metrics[0]

    # ASSERT that the parsed quality metrics are of the correct type
    assert isinstance(quality_metrics_model, BclConvertQualityMetrics)

    # ASSERT that the parsed quality metrics has the correct values
    for attr_name, attr_value in quality_metrics_model.dict().items():
        assert getattr(bcl_convert_quality_metric_model_with_data, attr_name) == attr_value


def test_parse_bcl_convert_adapter_metrics(
    parsed_bcl_convert_metrics: BclConvertMetricsParser,
    bcl_convert_adapter_metric_model_with_data: BclConvertAdapterMetrics,
):
    """Test to parse BCLConvert adapter metrics."""
    # GIVEN a parsed BCLConvert metrics

    # ASSERT that the parsed adapter metrics are correct
    adapter_metrics_model: BclConvertAdapterMetrics = parsed_bcl_convert_metrics.adapter_metrics[0]

    # ASSERT that the parsed adapter metrics are of the correct type
    assert isinstance(adapter_metrics_model, BclConvertAdapterMetrics)

    # ASSERT that the parsed adapter metrics has the correct values
    for attr_name, attr_value in adapter_metrics_model.dict().items():
        assert getattr(bcl_convert_adapter_metric_model_with_data, attr_name) == attr_value


def test_parse_bcl_convert_demux_metrics(
    parsed_bcl_convert_metrics: BclConvertMetricsParser,
    bcl_convert_demux_metric_model_with_data: BclConvertDemuxMetrics,
):
    """Test to parse BCLConvert demux metrics."""
    # GIVEN a parsed BCLConvert metrics

    # ASSERT that the parsed demux metrics are correct
    demux_metrics_model: BclConvertDemuxMetrics = parsed_bcl_convert_metrics.demux_metrics[0]

    # ASSERT that the parsed demux metrics are of the correct type
    assert isinstance(demux_metrics_model, BclConvertDemuxMetrics)

    # ASSERT that the parsed demux metrics has the correct values
    for attr_name, attr_value in demux_metrics_model.dict().items():
        assert getattr(bcl_convert_demux_metric_model_with_data, attr_name) == attr_value


def test_parse_bcl_convert_sample_sheet(
    parsed_bcl_convert_metrics: BclConvertMetricsParser,
    bcl_convert_sample_sheet_model_with_data: BclConvertSampleSheetData,
):
    """Test to parse BCLConvert sample sheet."""
    # GIVEN a parsed BCLConvert metrics

    # ASSERT that the parsed sample sheet is correct
    sample_sheet_model: BclConvertSampleSheetData = parsed_bcl_convert_metrics.sample_sheet[0]

    # ASSERT that the parsed sample sheet is of the correct type
    assert isinstance(sample_sheet_model, BclConvertSampleSheetData)

    # ASSERT that the parsed sample sheet has the correct values
    for attr_name, attr_value in sample_sheet_model.dict().items():
        assert getattr(bcl_convert_sample_sheet_model_with_data, attr_name) == attr_value



"""Module for reading and writing comma separated values (CSV) formatted files."""
import csv
import io
from typing import Any, List, Union
from pathlib import Path

from cg.constants import FileExtensions
from cg.io.validate_path import validate_file_suffix


def read_csv(file_path: Path, read_to_dict: bool = False) -> Union[List[List[str]], List[dict]]:
    """Read content in a CSV file to a list of list or list of dict."""
    validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtensions.CSV)
    with open(file_path, "r") as file:
        csv_reader = csv.DictReader(file) if read_to_dict else csv.reader(file)
        return list(csv_reader)


def read_csv_stream(stream: str) -> List[List[str]]:
    """Read CSV formatted stream."""
    csv_reader = csv.reader(stream.splitlines())
    return list(csv_reader)


def write_csv(content: List[List[Any]], file_path: Path) -> None:
    """Write content to a CSV file."""
    with open(file_path, "w", newline="") as file:
        csv_writer = csv.writer(file)
        for row in content:
            csv_writer.writerow(row)


def write_csv_stream(content: List[List[Any]]) -> str:
    """Write content to a CSV stream."""
    csv_stream = io.StringIO()
    csv_writer = csv.writer(csv_stream, lineterminator="\n")
    for row in content:
        csv_writer.writerow(row)
    return csv_stream.getvalue()



from cg.apps.cgstats.stats import StatsAPI

from cg.apps.cgstats.crud.delete import delete_flowcell


def test_delete_flowcell(populated_stats_api: StatsAPI, flow_cell_name: str):
    """Test to delete a flowcell from cg-stats"""

    # GIVEN a populated StatsAPI, with an existing flow cell
    flow_cell = populated_stats_api.find_handler.get_flow_cell_by_name(
        flow_cell_name=flow_cell_name
    )
    assert flow_cell

    # WHEN deleting a flow cell from the StatsAPI
    delete_flowcell(manager=populated_stats_api, flowcell_name=flow_cell_name)

    # THEN the flowcell should not exist any longer
    results = populated_stats_api.find_handler.get_flow_cell_by_name(flow_cell_name=flow_cell_name)
    assert not results



from pathlib import Path

from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.demultiplex.sample_sheet.models import SampleSheet
from cg.apps.demultiplex.sample_sheet.validate import validate_sample_sheet


def test_convert_to_bcl2fastq_sheet(
    novaseq_bcl2fastq_sample_sheet_object: SampleSheetCreator, project_dir: Path
):
    """Test that a created bcl2fastq sample sheet has samples."""
    # GIVEN a sample sheet object populated with samples
    assert novaseq_bcl2fastq_sample_sheet_object.lims_samples

    # WHEN converting to a sample sheet
    sample_sheet: str = novaseq_bcl2fastq_sample_sheet_object.construct_sample_sheet()

    # THEN assert a correctly formatted sample sheet was created
    sample_sheet_object: SampleSheet = validate_sample_sheet(
        sample_sheet_content=sample_sheet,
        bcl_converter=novaseq_bcl2fastq_sample_sheet_object.bcl_converter,
    )
    assert sample_sheet_object.samples


def test_convert_to_dragen_sheet(
    novaseq_dragen_sample_sheet_object: SampleSheetCreator, project_dir: Path
):
    """Test that a created bcl2fastq sample sheet has samples."""
    # GIVEN a sample sheet object populated with samples
    assert novaseq_dragen_sample_sheet_object.lims_samples

    # WHEN converting to a sample sheet
    sample_sheet: str = novaseq_dragen_sample_sheet_object.construct_sample_sheet()

    # THEN assert a correctly formatted sample sheet was created
    sample_sheet_object: SampleSheet = validate_sample_sheet(
        sample_sheet_content=sample_sheet,
        bcl_converter=novaseq_dragen_sample_sheet_object.bcl_converter,
    )
    assert sample_sheet_object.samples



from pathlib import Path

from cg.cli.demultiplex.add import add_flow_cell_cmd
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCell
from click.testing import CliRunner


def test_add_flowcell_cmd(
    cli_runner: CliRunner,
    bcl2fastq_flow_cell: FlowCell,
    demultiplex_context: CGConfig,
    demultiplexed_flow_cell_finished_working_directory: Path,
    demultiplex_ready_flow_cell: Path,
):
    # GIVEN a cgstats api and a demultiplex api
    # GIVEN that there is a flowcell in the run dir
    assert demultiplex_ready_flow_cell.exists()
    # GIVEN that there is a demultiplexed flowcell
    assert demultiplexed_flow_cell_finished_working_directory.exists()

    # GIVEN that the flowcell does not exist in the cgstats database
    assert not demultiplex_context.cg_stats_api.find_handler.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_flow_cell.id
    )

    # WHEN running the add flowcell command
    result = cli_runner.invoke(
        add_flow_cell_cmd, [bcl2fastq_flow_cell.full_name], obj=demultiplex_context
    )

    # THEN assert that the run was success
    assert result.exit_code == 0
    # THEN assert that the flowcell was added to cgstats
    assert demultiplex_context.cg_stats_api.find_handler.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_flow_cell.id
    )



from typing import Dict
import pytest


@pytest.fixture
def valid_bcl2fastq_metrics_data() -> Dict:
    """Example structure of a valid stats.json file containing sequecing metrics generated by bcl2fastq."""
    return {
        "Flowcell": "AB1",
        "RunNumber": 1,
        "RunId": "RUN1",
        "ReadInfosForLanes": [
            {
                "LaneNumber": 1,
            }
        ],
        "ConversionResults": [
            {
                "LaneNumber": 1,
                "TotalClustersRaw": 100,
                "TotalClustersPF": 100,
                "Yield": 1000,
                "DemuxResults": [
                    {
                        "SampleId": "S1",
                        "SampleName": "Sample1",
                        "IndexMetrics": [{"IndexSequence": "ATGC", "MismatchCounts": {"ATGC": 1}}],
                        "NumberReads": 1,
                        "Yield": 100,
                        "ReadMetrics": [
                            {"ReadNumber": 1, "Yield": 100, "YieldQ30": 90, "QualityScoreSum": 100}
                        ],
                    }
                ],
            }
        ],
    }



from typing import Optional

from pydantic import BaseModel, validator
from typing_extensions import Literal

from cg.constants import Priority

SEX_MAP = {"male": "M", "female": "F"}


class Udf(BaseModel):
    application: str
    capture_kit: Optional[str]
    collection_date: Optional[str]
    comment: Optional[str]
    concentration: Optional[str]
    concentration_sample: Optional[str]
    customer: str
    control: Optional[str]
    data_analysis: Optional[str]
    data_delivery: Optional[str]
    elution_buffer: Optional[str]
    extraction_method: Optional[str]
    family_name: str = "NA"
    formalin_fixation_time: Optional[str]
    index: Optional[str]
    index_number: Optional[str]
    lab_code: Optional[str]
    organism: Optional[str]
    organism_other: Optional[str]
    original_lab: Optional[str]
    original_lab_address: Optional[str]
    pool: Optional[str]
    post_formalin_fixation_time: Optional[str]
    pre_processing_method: Optional[str]
    primer: Optional[str]
    priority: str = Priority.standard.name
    quantity: Optional[str]
    reference_genome: Optional[str]
    region: Optional[str]
    region_code: Optional[str]
    require_qc_ok: bool = False
    rml_plate_name: Optional[str]
    selection_criteria: Optional[str]
    sex: Literal["M", "F", "unknown"] = "unknown"
    skip_reception_control: Optional[bool] = None
    source: str = "NA"
    tissue_block_size: Optional[str]
    tumour: Optional[bool] = False
    tumour_purity: Optional[str]
    volume: Optional[str]
    well_position_rml: Optional[str]
    verified_organism: Optional[bool]

    @validator("sex", pre=True)
    def validate_sex(cls, value: str):
        return SEX_MAP.get(value, "unknown")


class LimsSample(BaseModel):
    name: str
    container: str = "Tube"
    container_name: Optional[str]
    well_position: Optional[str]
    index_sequence: Optional[str]
    udfs: Optional[Udf]

    @classmethod
    def parse_obj(cls, obj: dict):
        parsed_obj: LimsSample = super().parse_obj(obj)
        udf: Udf = Udf.parse_obj(obj)
        parsed_obj.udfs = udf
        return parsed_obj



import logging

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner
from cg.cli.workflow.taxprofiler.base import (
    taxprofiler,
)
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_taxprofiler_no_args(cli_runner: CliRunner, taxprofiler_context: CGConfig):
    """Test to see that running Taxprofiler without options prints help and doesn't result in an error."""
    # GIVEN no arguments or options besides the command call

    # WHEN running command
    result = cli_runner.invoke(taxprofiler, [], obj=taxprofiler_context)

    assert result.exit_code == EXIT_SUCCESS

    ## THEN help should be printed
    assert "help" in result.output



"""Tests the findbusinessdata part of the Cg store API."""
import logging
from datetime import datetime, timedelta
from typing import List

import pytest
from cg.constants import FlowCellStatus
from cg.constants.constants import CaseActions
from cg.constants.indexes import ListIndexes
from cg.exc import CgError
from cg.store import Store
from cg.store.models import (
    Application,
    ApplicationVersion,
    Customer,
    Family,
    FamilySample,
    Flowcell,
    Invoice,
    Pool,
    Sample,
)
from sqlalchemy.orm import Query
from tests.store_helpers import StoreHelpers


def test_get_analysis_by_case_entry_id_and_started_at(
    sample_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test returning an analysis using a date."""
    # GIVEN a case with an analysis with a start date in the database
    analysis = helpers.add_analysis(store=sample_store, started_at=timestamp_now)
    assert analysis.started_at

    # WHEN getting analysis via case_id and start date
    db_analysis = sample_store.get_analysis_by_case_entry_id_and_started_at(
        case_entry_id=analysis.family.id, started_at_date=analysis.started_at
    )

    # THEN the analysis should have been retrieved
    assert db_analysis == analysis


def test_get_flow_cell_sample_links_query(re_sequenced_sample_store: Store):
    """Test function to return the flow cell sample links query from the database."""

    # GIVEN a store with two flow cells

    # WHEN getting the query for the flow cells
    flow_cell_query: Query = re_sequenced_sample_store._get_join_flow_cell_sample_links_query()

    # THEN a query should be returned
    assert isinstance(flow_cell_query, Query)


def test_get_flow_cells(re_sequenced_sample_store: Store):
    """Test function to return the flow cells from the database."""

    # GIVEN a store with two flow cells

    # WHEN fetching the flow cells
    flow_cells: List[Flowcell] = re_sequenced_sample_store._get_query(table=Flowcell)

    # THEN a flow cells should be returned
    assert flow_cells

    # THEN a flow cell model should be returned
    assert isinstance(flow_cells[0], Flowcell)


def test_get_flow_cell(bcl2fastq_flow_cell_id: str, re_sequenced_sample_store: Store):
    """Test returning the latest flow cell from the database."""

    # GIVEN a store with two flow cells

    # WHEN fetching the latest flow cell
    flow_cell: Flowcell = re_sequenced_sample_store.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_flow_cell_id
    )

    # THEN the returned flow cell should have the same name as the one in the database
    assert flow_cell.name == bcl2fastq_flow_cell_id


def test_get_flow_cell_by_name_pattern(
    bcl2fastq_flow_cell_id: str, re_sequenced_sample_store: Store
):
    """Test returning the latest flow cell from the database by enquiry."""

    # GIVEN a store with two flow cells

    # WHEN fetching the latest flow cell
    flow_cell: List[Flowcell] = re_sequenced_sample_store.get_flow_cell_by_name_pattern(
        name_pattern=bcl2fastq_flow_cell_id[:4]
    )

    # THEN the returned flow cell should have the same name as the one in the database
    assert flow_cell.name == bcl2fastq_flow_cell_id


def test_get_flow_cells_by_case(
    base_store: Store,
    bcl2fastq_flow_cell_id: str,
    another_flow_cell_id: str,
    case: Family,
    helpers: StoreHelpers,
    sample: Sample,
):
    """Test returning the latest flow cell from the database by case."""

    # GIVEN a store with two flow cell
    helpers.add_flowcell(store=base_store, flow_cell_name=bcl2fastq_flow_cell_id, samples=[sample])

    helpers.add_flowcell(store=base_store, flow_cell_name=another_flow_cell_id)

    # WHEN fetching the latest flow cell
    flow_cells: List[Flowcell] = base_store.get_flow_cells_by_case(case=case)

    # THEN the flow cell samples for the case should be returned
    for flow_cell in flow_cells:
        for sample in flow_cell.samples:
            assert sample in case.samples

    # THEN the returned flow cell should have the same name as the one in the database
    assert flow_cells[0].name == bcl2fastq_flow_cell_id


def test_get_flow_cells_by_statuses(another_flow_cell_id: str, re_sequenced_sample_store: Store):
    """Test returning the latest flow cell from the database by statuses."""

    # GIVEN a store with two flow cells

    # WHEN fetching the latest flow cell
    flow_cells: List[Flowcell] = re_sequenced_sample_store.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.ON_DISK, FlowCellStatus.REQUESTED]
    )

    # THEN the flow cell status should be "ondisk"
    for flow_cell in flow_cells:
        assert flow_cell.status == FlowCellStatus.ON_DISK

    # THEN the returned flow cell should have the same name as the one in the database
    assert flow_cells[0].name == another_flow_cell_id


def test_get_flow_cells_by_statuses_when_multiple_matches(
    another_flow_cell_id: str, re_sequenced_sample_store: Store
):
    """Test returning the latest flow cell from the database by statuses when multiple matches."""

    # GIVEN a store with two flow cells

    # GIVEN a flow cell that exist in status db with status "requested"
    flow_cells: List[Flowcell] = re_sequenced_sample_store._get_query(table=Flowcell)
    flow_cells[0].status = FlowCellStatus.REQUESTED
    re_sequenced_sample_store.session.add(flow_cells[0])
    re_sequenced_sample_store.session.commit()

    # WHEN fetching the latest flow cell
    flow_cells: List[Flowcell] = re_sequenced_sample_store.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.ON_DISK, FlowCellStatus.REQUESTED]
    )

    # THEN the flow cell status should be "ondisk" or "requested"
    for flow_cell in flow_cells:
        assert flow_cell.status in [FlowCellStatus.ON_DISK, FlowCellStatus.REQUESTED]

    # THEN the returned flow cell should have the same status as the ones in the database
    assert flow_cells[0].status == FlowCellStatus.REQUESTED

    assert flow_cells[1].status == FlowCellStatus.ON_DISK


def test_get_flow_cells_by_statuses_when_incorrect_status(re_sequenced_sample_store: Store):
    """Test returning the latest flow cell from the database when no flow cell with incorrect status."""

    # GIVEN a store with two flow cells

    # WHEN fetching the latest flow cell
    flow_cells: List[Flowcell] = re_sequenced_sample_store.get_flow_cells_by_statuses(
        flow_cell_statuses=["does_not_exist"]
    )

    # THEN no flow cells should be returned
    assert not list(flow_cells)


def test_get_flow_cell_by_enquiry_and_status(
    bcl2fastq_flow_cell_id: str, re_sequenced_sample_store: Store
):
    """Test returning the latest flow cell from the database by enquiry and status."""

    # GIVEN a store with two flow cells

    # WHEN fetching the latest flow cell
    flow_cell: List[Flowcell] = re_sequenced_sample_store.get_flow_cell_by_name_pattern_and_status(
        flow_cell_statuses=[FlowCellStatus.ON_DISK], name_pattern=bcl2fastq_flow_cell_id[:4]
    )

    # THEN the returned flow cell should have the same name as the one in the database
    assert flow_cell[0].name == bcl2fastq_flow_cell_id

    # THEN the returned flow cell should have the same status as the query
    assert flow_cell[0].status == FlowCellStatus.ON_DISK


def test_get_samples_from_flow_cell(
    bcl2fastq_flow_cell_id: str, sample_id: str, re_sequenced_sample_store: Store
):
    """Test returning samples present on the latest flow cell from the database."""

    # GIVEN a store with two flow cells

    # WHEN fetching the samples from the latest flow cell
    samples: List[Sample] = re_sequenced_sample_store.get_samples_from_flow_cell(
        flow_cell_id=bcl2fastq_flow_cell_id
    )

    # THEN the returned sample id should have the same id as the one in the database
    assert samples[0].internal_id == sample_id


def test_get_latest_flow_cell_on_case(
    re_sequenced_sample_store: Store, case_id: str, bcl2fastq_flow_cell_id: str
):
    """Test returning the latest sequenced flow cell on a case."""

    # GIVEN a store with two flow cells in it, one being the latest sequenced of the two
    latest_flow_cell: Flowcell = re_sequenced_sample_store.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_flow_cell_id
    )

    # WHEN fetching the latest flow cell on a case with a sample that has been sequenced on both flow cells
    latest_flow_cell_on_case: Flowcell = re_sequenced_sample_store.get_latest_flow_cell_on_case(
        family_id=case_id
    )

    # THEN the fetched flow cell should have the same name as the other
    assert latest_flow_cell.name == latest_flow_cell_on_case.name


def test_is_all_flow_cells_on_disk_when_no_flow_cell(
    base_store: Store,
    caplog,
    case_id: str,
):
    """Test check if all flow cells for samples on a case is on disk when no flow cells."""
    caplog.set_level(logging.DEBUG)

    # WHEN fetching the latest flow cell
    is_on_disk = base_store.is_all_flow_cells_on_disk(case_id=case_id)

    # THEN return false
    assert is_on_disk is False

    # THEN log no flow cells found
    assert "No flow cells found" in caplog.text


def test_is_all_flow_cells_on_disk_when_not_on_disk(
    base_store: Store,
    caplog,
    bcl2fastq_flow_cell_id: str,
    another_flow_cell_id: str,
    case_id: str,
    helpers: StoreHelpers,
    sample: Sample,
):
    """Test check if all flow cells for samples on a case is on disk when not on disk."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a store with two flow cell
    flow_cell = helpers.add_flowcell(
        store=base_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        samples=[sample],
        status=FlowCellStatus.PROCESSING,
    )

    another_flow_cell = helpers.add_flowcell(
        store=base_store,
        flow_cell_name=another_flow_cell_id,
        samples=[sample],
        status=FlowCellStatus.RETRIEVED,
    )

    # WHEN fetching the latest flow cell
    is_on_disk = base_store.is_all_flow_cells_on_disk(case_id=case_id)

    # THEN return false
    assert is_on_disk is False

    # THEN log the status of the flow cell
    assert f"{flow_cell.name}: {flow_cell.status}" in caplog.text
    assert f"{another_flow_cell.name}: {another_flow_cell.status}" in caplog.text


def test_is_all_flow_cells_on_disk_when_requested(
    base_store: Store,
    caplog,
    bcl2fastq_flow_cell_id: str,
    another_flow_cell_id: str,
    case_id: str,
    helpers: StoreHelpers,
    sample: Sample,
):
    """Test check if all flow cells for samples on a case is on disk when requested."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a store with two flow cell
    flow_cell = helpers.add_flowcell(
        store=base_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        samples=[sample],
        status=FlowCellStatus.REMOVED,
    )

    another_flow_cell = helpers.add_flowcell(
        store=base_store,
        flow_cell_name=another_flow_cell_id,
        samples=[sample],
        status=FlowCellStatus.REQUESTED,
    )

    # WHEN fetching the latest flow cell
    is_on_disk = base_store.is_all_flow_cells_on_disk(case_id=case_id)

    # THEN return false
    assert is_on_disk is False

    # THEN log the requesting the flow cell
    assert f"{flow_cell.name}: flow cell not on disk, requesting" in caplog.text

    # THEN log the status of the flow cell
    assert f"{another_flow_cell.name}: {another_flow_cell.status}" in caplog.text


def test_is_all_flow_cells_on_disk(
    base_store: Store,
    caplog,
    bcl2fastq_flow_cell_id: str,
    another_flow_cell_id: str,
    case_id: str,
    helpers: StoreHelpers,
    sample: Sample,
):
    """Test check if all flow cells for samples on a case is on disk."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a store with two flow cell
    flow_cell = helpers.add_flowcell(
        store=base_store, flow_cell_name=bcl2fastq_flow_cell_id, samples=[sample]
    )

    helpers.add_flowcell(store=base_store, flow_cell_name=another_flow_cell_id)

    # WHEN fetching the latest flow cell
    is_on_disk = base_store.is_all_flow_cells_on_disk(case_id=case_id)

    # THEN return true
    assert is_on_disk is True

    # THEN log the status of the flow cell
    assert f"{flow_cell.name}: status is {flow_cell.status}" in caplog.text


def test_get_customer_id_from_ticket(analysis_store, customer_id, ticket_id: str):
    """Tests if the function in fact returns the correct customer."""
    # Given a store with a ticket

    # Then the function should return the customer connected to the ticket
    assert analysis_store.get_customer_id_from_ticket(ticket_id) == customer_id


def test_get_latest_ticket_from_case(case_id: str, analysis_store_single_case, ticket_id: str):
    """Tests if the correct ticket is returned for the given case."""
    # GIVEN a populated store with a case

    # WHEN the function is called
    ticket_from_case: str = analysis_store_single_case.get_latest_ticket_from_case(case_id=case_id)

    # THEN the ticket should be correct
    assert ticket_id == ticket_from_case


def test_get_ready_made_library_expected_reads(case_id: str, rml_pool_store: Store):
    """Test if the correct number of expected reads is returned."""

    # GIVEN a case with a sample with an application version
    application_version: ApplicationVersion = (
        rml_pool_store.get_case_by_internal_id(internal_id=case_id)
        .links[ListIndexes.FIRST.value]
        .sample.application_version
    )

    # WHEN the expected reads is fetched from the case
    expected_reads: int = rml_pool_store.get_ready_made_library_expected_reads(case_id=case_id)

    # THEN the fetched reads should be equal to the expected reads of the application versions application
    assert application_version.application.expected_reads == expected_reads


def test_get_application_by_case(case_id: str, rml_pool_store: Store):
    """Test that the correct application is returned on a case."""
    # GIVEN a case with a sample with an application version
    application_version: ApplicationVersion = (
        rml_pool_store.get_case_by_internal_id(internal_id=case_id)
        .links[ListIndexes.FIRST.value]
        .sample.application_version
    )

    # WHEN the application is fetched from the case
    application: Application = rml_pool_store.get_application_by_case(case_id=case_id)

    # THEN the fetched application should be equal to the application version application
    assert application_version.application == application


def test_get_case_samples_by_case_id(
    store_with_analyses_for_cases: Store,
    case_id: str,
):
    """Test that getting case-samples by case id returns a list of FamilySamples."""
    # GIVEN a store with case-samples and a case id

    # WHEN fetching the case-samples matching the case id
    case_samples: List[FamilySample] = store_with_analyses_for_cases.get_case_samples_by_case_id(
        case_internal_id=case_id
    )

    # THEN a list of case-samples should be returned
    assert case_samples
    assert isinstance(case_samples, List)
    assert isinstance(case_samples[0], FamilySample)


def test_get_case_sample_link(
    store_with_analyses_for_cases: Store,
    case_id: str,
    sample_id: str,
):
    """Test that the returned element is a FamilySample with the correct case and sample internal ids."""
    # GIVEN a store with case-samples and valid case and sample internal ids

    # WHEN fetching a case-sample with case and sample internal ids
    case_sample: FamilySample = store_with_analyses_for_cases.get_case_sample_link(
        case_internal_id=case_id,
        sample_internal_id=sample_id,
    )

    # THEN the returned element is a FamilySample object
    assert isinstance(case_sample, FamilySample)
    # THEN the returned family sample has the correct case and sample internal ids
    assert case_sample.family.internal_id == case_id
    assert case_sample.sample.internal_id == sample_id


def test_find_single_case_for_sample(
    sample_id_in_single_case: str, store_with_multiple_cases_and_samples: Store
):
    """Test that cases associated with a sample can be found."""

    # GIVEN a database containing a sample associated with a single case
    sample: Sample = store_with_multiple_cases_and_samples.get_sample_by_internal_id(
        internal_id=sample_id_in_single_case
    )

    assert sample

    # WHEN the cases associated with the sample is fetched
    cases: List[
        FamilySample
    ] = store_with_multiple_cases_and_samples.get_case_samples_from_sample_entry_id(
        sample_entry_id=sample.id
    ).all()

    # THEN only one case is found
    assert cases and len(cases) == 1


def test_find_multiple_cases_for_sample(
    sample_id_in_multiple_cases: str, store_with_multiple_cases_and_samples: Store
):
    # GIVEN a database containing a sample associated with multiple cases
    sample: Sample = store_with_multiple_cases_and_samples.get_sample_by_internal_id(
        internal_id=sample_id_in_multiple_cases
    )
    assert sample

    # WHEN the cases associated with the sample is fetched
    cases: List[
        FamilySample
    ] = store_with_multiple_cases_and_samples.get_case_samples_from_sample_entry_id(
        sample_entry_id=sample.id
    ).all()

    # THEN multiple cases are found
    assert cases and len(cases) > 1


def test_find_cases_for_non_existing_case(store_with_multiple_cases_and_samples: Store):
    """Test that nothing happens when trying to find a case that does not exist."""

    # GIVEN a database containing some cases but not a specific case
    case_id: str = "some_case"
    case: Family = store_with_multiple_cases_and_samples.get_case_by_internal_id(
        internal_id=case_id
    )

    assert not case

    # WHEN trying to find cases with samples given the non existing case id
    cases = store_with_multiple_cases_and_samples.filter_cases_with_samples(case_ids=[case_id])

    # THEN no cases are found
    assert not cases


def test_verify_case_exists(
    caplog, case_id_with_multiple_samples: str, store_with_multiple_cases_and_samples: Store
):
    """Test validating a case that exists in the database."""

    caplog.set_level(logging.INFO)

    # GIVEN a database containing the case

    # WHEN validating if the case exists
    store_with_multiple_cases_and_samples.verify_case_exists(
        case_internal_id=case_id_with_multiple_samples
    )

    # THEN the case is found
    assert f"Case {case_id_with_multiple_samples} exists in Status DB" in caplog.text


def test_verify_case_exists_with_non_existing_case(
    caplog, case_id_does_not_exist: str, store_with_multiple_cases_and_samples: Store
):
    """Test validating a case that does not exist in the database."""

    # GIVEN a database containing the case

    with pytest.raises(CgError):
        # WHEN validating if the case exists
        store_with_multiple_cases_and_samples.verify_case_exists(
            case_internal_id=case_id_does_not_exist
        )

        # THEN the case is not found
        assert f"Case {case_id_does_not_exist} could not be found in Status DB!" in caplog.text


def test_verify_case_exists_with_no_case_samples(
    caplog, case_id_without_samples: str, store_with_multiple_cases_and_samples: Store
):
    """Test validating a case without samples that exist in the database."""

    # GIVEN a database containing the case

    with pytest.raises(CgError):
        # WHEN validating if the case exists
        store_with_multiple_cases_and_samples.verify_case_exists(
            case_internal_id=case_id_without_samples
        )

        # THEN the case is found, but has no samples
        assert "Case {case_id} has no samples in in Status DB!" in caplog.text


def test_is_case_down_sampled_true(base_store: Store, case: Family, sample_id: str):
    """Tests the down sampling check when all samples are down sampled."""
    # GIVEN a case where all samples are down sampled
    for sample in case.samples:
        sample.from_sample = sample_id
    base_store.session.commit()

    # WHEN checking if all sample in the case are down sampled
    is_down_sampled: bool = base_store.is_case_down_sampled(case_id=case.internal_id)

    # THEN the return value should be True
    assert is_down_sampled


def test_is_case_down_sampled_false(base_store: Store, case: Family, sample_id: str):
    """Tests the down sampling check when none of the samples are down sampled."""
    # GIVEN a case where all samples are not down sampled
    for sample in case.samples:
        assert not sample.from_sample

    # WHEN checking if all sample in the case are down sampled
    is_down_sampled: bool = base_store.is_case_down_sampled(case_id=case.internal_id)

    # THEN the return value should be False
    assert not is_down_sampled


def test_is_case_external_true(
    base_store: Store, case: Family, helpers: StoreHelpers, sample_id: str
):
    """Tests the external case check when all the samples are external."""
    # GIVEN a case where all samples are not external
    external_application_version: ApplicationVersion = helpers.ensure_application_version(
        store=base_store, is_external=True
    )
    for sample in case.samples:
        sample.application_version = external_application_version
    base_store.session.commit()

    # WHEN checking if all sample in the case are external
    is_external: bool = base_store.is_case_external(case_id=case.internal_id)

    # THEN the return value should be False
    assert is_external


def test_is_case_external_false(base_store: Store, case: Family, sample_id: str):
    """Tests the external case check when none of the samples are external."""
    # GIVEN a case where all samples are not external
    for sample in case.samples:
        assert not sample.application_version.application.is_external

    # WHEN checking if all sample in the case are external
    is_external: bool = base_store.is_case_external(case_id=case.internal_id)

    # THEN the return value should be False
    assert not is_external


def test_get_invoice_by_status(store_with_an_invoice_with_and_without_attributes: Store):
    """Test that invoices can be fetched by status."""
    # GIVEN a database with two invoices of which one has attributes

    # WHEN fetching the invoice by status
    invoices: List[
        Invoice
    ] = store_with_an_invoice_with_and_without_attributes.get_invoices_by_status(is_invoiced=True)

    # THEN one invoice should be returned
    assert invoices

    # THEN there should be one invoice
    assert len(invoices) == 1

    # THEN the invoice should have that is invoiced
    assert invoices[0].invoiced_at


def test_get_invoice_by_id(store_with_an_invoice_with_and_without_attributes: Store):
    """Test that invoices can be fetched by invoice id."""
    # GIVEN a database with two invoices of which one has attributes

    # WHEN fetching the invoice by invoice id
    invoice: Invoice = store_with_an_invoice_with_and_without_attributes.get_invoice_by_entry_id(
        entry_id=1
    )

    # THEN one invoice should be returned
    assert invoice

    # THEN the invoice should have id 1
    assert invoice.id == 1


def test_get_pools(store_with_multiple_pools_for_customer: Store):
    """Test that pools can be fetched from the store."""
    # GIVEN a database with two pools

    # WHEN getting all pools
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools()

    # THEN two pools should be returned
    assert len(pools) == 2


def test_get_pools_by_customer_id(store_with_multiple_pools_for_customer: Store):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools

    # WHEN getting pools by customer id
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_by_customer_id(
        customers=store_with_multiple_pools_for_customer.get_customers()
    )

    # THEN two pools should be returned
    assert len(pools) == 2


def test_get_pools_by_name_enquiry(store_with_multiple_pools_for_customer: Store, pool_name_1: str):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools

    # WHEN fetching pools by customer id
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_by_name_enquiry(
        name_enquiry=pool_name_1
    )

    # THEN one pool should be returned
    assert len(pools) == 1


def test_get_pools_by_order_enquiry(
    store_with_multiple_pools_for_customer: Store, pool_order_1: str
):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools

    # WHEN getting pools by customer id
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_by_order_enquiry(
        order_enquiry=pool_order_1
    )

    # THEN one pool should be returned
    assert len(pools) == 1


def test_get_pools_to_render_with(
    store_with_multiple_pools_for_customer: Store,
    pool_name_1: str,
    pool_order_1: str,
):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools

    # WHEN fetching pools with no customer or enquiry
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_to_render()

    # THEN two pools should be returned
    assert len(pools) == 2


def test_get_pools_to_render_with_customer(
    store_with_multiple_pools_for_customer: Store,
):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools

    # WHEN getting pools by customer id
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_to_render(
        customers=store_with_multiple_pools_for_customer.get_customers()
    )

    # THEN two pools should be returned
    assert len(pools) == 2


def test_get_pools_to_render_with_customer_and_name_enquiry(
    store_with_multiple_pools_for_customer: Store,
    pool_name_1: str,
):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools
    # WHEN fetching pools by customer id and name enquiry
    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_to_render(
        customers=store_with_multiple_pools_for_customer.get_customers(), enquiry=pool_name_1
    )

    # THEN one pools should be returned
    assert len(pools) == 1


def test_get_pools_to_render_with_customer_and_order_enquiry(
    store_with_multiple_pools_for_customer: Store,
    pool_order_1: str,
):
    """Test that pools can be fetched from the store by customer id."""
    # GIVEN a database with two pools

    # WHEN fetching pools by customer id and order enquiry

    pools: List[Pool] = store_with_multiple_pools_for_customer.get_pools_to_render(
        customers=store_with_multiple_pools_for_customer.get_customers(), enquiry=pool_order_1
    )

    # THEN one pools should be returned
    assert len(pools) == 1


def test_get_case_by_name_and_customer_case_found(store_with_multiple_cases_and_samples: Store):
    """Test that a case can be found by customer and case name."""
    # GIVEN a database with multiple cases for a customer
    case: Family = store_with_multiple_cases_and_samples._get_query(table=Family).first()
    customer: Customer = store_with_multiple_cases_and_samples._get_query(table=Customer).first()

    assert case.customer == customer

    # WHEN fetching a case by customer and case name
    filtered_case: Family = store_with_multiple_cases_and_samples.get_case_by_name_and_customer(
        customer=customer,
        case_name=case.name,
    )

    # THEN the correct case should be returned
    assert filtered_case is not None
    assert filtered_case.customer_id == customer.id
    assert filtered_case.name == case.name


def test_get_cases_not_analysed_by_sample_internal_id_empty_query(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    """Test that an empty query is returned if no cases match the sample internal id."""
    # GIVEN a store
    # WHEN getting cases not analysed by the sample internal id
    cases = store_with_multiple_cases_and_samples.get_not_analysed_cases_by_sample_internal_id(
        sample_internal_id=non_existent_id
    )

    # THEN an empty list should be returned
    assert cases == []


def test_get_cases_not_analysed_by_sample_internal_id_multiple_cases(
    store_with_multiple_cases_and_samples: Store,
    sample_id_in_multiple_cases: str,
):
    """Test that multiple cases are returned when more than one case matches the sample internal id."""
    # GIVEN a store with multiple cases having the same sample internal id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Family)

    # Set all cases to not analysed and HOLD action
    for case in cases_query.all():
        case.action = CaseActions.HOLD

    # WHEN getting cases not analysed by the shared sample internal id
    cases = store_with_multiple_cases_and_samples.get_not_analysed_cases_by_sample_internal_id(
        sample_id_in_multiple_cases
    )

    # THEN multiple cases should be returned
    assert len(cases) > 1

    # Check that all returned cases have the matching sample and are not analysed
    for case in cases:
        assert not case.analyses
        assert any(sample.internal_id == sample_id_in_multiple_cases for sample in case.samples)


def test_fetch_cases_newer_than_date_no_cases(store_with_multiple_cases_and_samples: Store):
    """Test that no cases are returned when there are no cases newer than the given date."""
    # GIVEN a store with cases older than 7 days
    older_than_date = datetime.now() - timedelta(days=10)
    for case in store_with_multiple_cases_and_samples._get_query(table=Family):
        case.created_at = older_than_date

    # WHEN fetching cases newer than 7 days
    cases = store_with_multiple_cases_and_samples.get_cases_created_within_days(days=7)

    # THEN no cases should be returned
    assert len(cases) == 0


def test_fetch_cases_newer_than_date_all_cases(store_with_multiple_cases_and_samples: Store):
    """Test that all cases are returned when all cases newer than the given date."""
    # GIVEN a store with cases newer than 7 days
    older_than_date = datetime.now() - timedelta(days=5)
    all_cases = store_with_multiple_cases_and_samples._get_query(table=Family).all()
    for case in all_cases:
        case.created_at = older_than_date

    # WHEN fetching cases newer than 7 days
    cases = store_with_multiple_cases_and_samples.get_cases_created_within_days(days=7)

    # THEN all cases should be returned
    assert len(cases) == len(all_cases)



"""Constants for cg."""

import click

from cg.utils.enums import StrEnum
from cg.utils.date import get_date

VALID_DATA_IN_PRODUCTION = get_date("2017-09-27")

MAX_ITEMS_TO_RETRIEVE = 50

SCALE_TO_MILLION_READ_PAIRS = 2_000_000

ANALYSIS_TYPES = ["tumor_wgs", "tumor_normal_wgs", "tumor_panel", "tumor_normal_panel"]

CAPTUREKIT_CANCER_OPTIONS = (
    "GIcfDNA",
    "GMCKsolid",
    "GMSmyeloid",
    "LymphoMATIC",
    "other (specify in comment field)",
)
CAPTUREKIT_OPTIONS = (
    "Agilent Sureselect CRE",
    "Agilent Sureselect V5",
    "SureSelect Focused Exome",
    "Twist_Target_hg19.bed",
    "other",
)


class CaseActions(StrEnum):
    ANALYZE: str = "analyze"
    HOLD: str = "hold"
    RUNNING: str = "running"


CASE_ACTIONS = [action.value for action in CaseActions]

COLLABORATORS = ("cust000", "cust002", "cust003", "cust004", "cust042")

COMBOS = {
    "DSD": ("DSD", "DSD-S", "HYP", "SEXDIF", "SEXDET"),
    "CM": ("CNM", "CM"),
    "Horsel": ("Horsel", "141217", "141201"),
}

CONTAINER_OPTIONS = ("Tube", "96 well plate", "No container")

CONTROL_OPTIONS = ("", "negative", "positive")

DEFAULT_CAPTURE_KIT = "twistexomerefseq_9.1_hg19_design.bed"


class FlowCellStatus(StrEnum):
    ON_DISK: str = "ondisk"
    REMOVED: str = "removed"
    REQUESTED: str = "requested"
    PROCESSING: str = "processing"
    RETRIEVED: str = "retrieved"


FLOWCELL_STATUS = [status.value for status in FlowCellStatus]


class AnalysisType(StrEnum):
    TARGETED_GENOME_SEQUENCING: str = "tgs"
    WHOLE_EXOME_SEQUENCING: str = "wes"
    WHOLE_GENOME_SEQUENCING: str = "wgs"
    WHOLE_TRANSCRIPTOME_SEQUENCING: str = "wts"
    OTHER: str = "other"


class PrepCategory(StrEnum):
    COVID: str = "cov"
    MICROBIAL: str = "mic"
    READY_MADE_LIBRARY: str = "rml"
    TARGETED_GENOME_SEQUENCING: str = "tgs"
    WHOLE_EXOME_SEQUENCING: str = "wes"
    WHOLE_GENOME_SEQUENCING: str = "wgs"
    WHOLE_TRANSCRIPTOME_SEQUENCING: str = "wts"


PREP_CATEGORIES = ("cov", "mic", "rml", "tgs", "wes", "wgs", "wts")

SEX_OPTIONS = ("male", "female", "unknown")

SARS_COV_REGEX = "^[0-9]{2}CS[0-9]{6}$"

STATUS_OPTIONS = ("affected", "unaffected", "unknown")


class FileFormat(StrEnum):
    FASTQ: str = "fastq"
    JSON: str = "json"
    YAML: str = "yaml"
    CSV: str = "csv"


class GenomeVersion(StrEnum):
    hg19: str = "hg19"
    hg38: str = "hg38"
    canfam3: str = "canfam3"


class SampleType(StrEnum):
    TUMOR: str = "tumor"
    NORMAL: str = "normal"


class DataDelivery(StrEnum):
    ANALYSIS_FILES: str = "analysis"
    ANALYSIS_SCOUT: str = "analysis-scout"
    FASTQ: str = "fastq"
    FASTQ_SCOUT: str = "fastq-scout"
    FASTQ_QC: str = "fastq_qc"
    FASTQ_ANALYSIS: str = "fastq-analysis"
    FASTQ_QC_ANALYSIS: str = "fastq_qc-analysis"
    FASTQ_ANALYSIS_SCOUT: str = "fastq-analysis-scout"
    NIPT_VIEWER: str = "nipt-viewer"
    NO_DELIVERY: str = "no-delivery"
    SCOUT: str = "scout"
    STATINA: str = "statina"


class HastaSlurmPartitions(StrEnum):
    DRAGEN: str = "dragen"


class FileExtensions(StrEnum):
    BED: str = ".bed"
    CRAM: str = ".cram"
    CSV: str = ".csv"
    GPG: str = ".gpg"
    GZIP: str = ".gz"
    JSON: str = ".json"
    KEY: str = ".key"
    NO_EXTENSION: str = ""
    SPRING: str = ".spring"
    TAR: str = ".tar"
    TMP: str = ".tmp"
    VCF: str = ".vcf"


class APIMethods(StrEnum):
    POST: str = "POST"
    PUT: str = "PUT"
    GET: str = "GET"
    DELETE: str = "DELETE"
    PATCH: str = "PATCH"


DRY_RUN = click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    default=False,
    help="Runs the command without making any changes",
)

SKIP_CONFIRMATION = click.option(
    "-y",
    "--yes",
    is_flag=True,
    default=False,
    help="Skip confirmation",
)


class MicrosaltQC:
    QC_PERCENT_THRESHOLD_MWX: float = 0.1
    COVERAGE_10X_THRESHOLD: float = 0.75
    NEGATIVE_CONTROL_READS_THRESHOLD: float = 0.2
    TARGET_READS: int = 6000000


class MicrosaltAppTags(StrEnum):
    MWRNXTR003: str = "MWRNXTR003"
    MWXNXTR003: str = "MWXNXTR003"
    PREP_CATEGORY: str = "mic"


DRY_RUN_MESSAGE = "Dry run: process call will not be executed!"


class MetaApis:
    ANALYSIS_API: str = "analysis_api"


class WorkflowManager(StrEnum):
    Slurm: str = "slurm"
    Tower: str = "nf_tower"



import logging
from pathlib import Path
from typing import Dict, Iterator, List, Union

import alchy
import sqlalchemy as sqa

from cg.apps.cgstats.crud.find import FindHandler
from cg.apps.cgstats.db.models import (
    Datasource,
    Demux,
    Flowcell,
    Model,
    Project,
    Sample,
    Unaligned,
    Supportparams,
)
from cg.constants import FLOWCELL_Q30_THRESHOLD
from cg.models.cgstats.flowcell import StatsFlowcell, StatsSample

LOG = logging.getLogger(__name__)


class StatsAPI(alchy.Manager):
    Project = Project
    Sample = Sample
    Unaligned = Unaligned
    Supportparams = Supportparams
    Datasource = Datasource
    Demux = Demux
    Flowcell = Flowcell

    def __init__(self, config: dict):
        LOG.info("Instantiating cgstats api")
        alchy_config = dict(SQLALCHEMY_DATABASE_URI=config["cgstats"]["database"])
        super(StatsAPI, self).__init__(config=alchy_config, Model=Model)
        self.root_dir: Path = Path(config["cgstats"]["root"])
        self.binary: str = config["cgstats"]["binary_path"]
        self.db_uri: str = config["cgstats"]["database"]
        self.find_handler = FindHandler()

    @staticmethod
    def get_curated_sample_name(sample_name: str) -> str:
        """Create new sample name"""
        raw_sample_name: str = sample_name.split("_", 1)[0]
        return raw_sample_name.rstrip("AB")

    def get_flowcell_samples(self, flowcell_object: Flowcell) -> List[StatsSample]:
        flowcell_samples: List[StatsSample] = []
        for sample_obj in self.flowcell_samples(flowcell_obj=flowcell_object):
            curated_sample_name: str = self.get_curated_sample_name(sample_obj.samplename)
            sample_data = {"name": curated_sample_name, "reads": 0, "fastqs": []}
            for fc_data in self.sample_reads(sample_obj):
                if fc_data.q30 >= FLOWCELL_Q30_THRESHOLD[fc_data.type]:
                    sample_data["reads"] += fc_data.reads
                else:
                    q30_threshold: int = FLOWCELL_Q30_THRESHOLD[fc_data.type]
                    LOG.warning(
                        f"q30 too low for {curated_sample_name} on {fc_data.name}:"
                        f"{fc_data.q30} < {q30_threshold}%"
                    )
                    continue

                for fastq_path in self.fastqs(fc_data.name, sample_obj):
                    if self.is_lane_pooled(
                        flowcell_obj=flowcell_object, lane=fc_data.lane
                    ) and "Undetermined" in str(fastq_path):
                        continue
                    sample_data["fastqs"].append(str(fastq_path))
                flowcell_samples.append(StatsSample(**sample_data))
        return flowcell_samples

    def flowcell(self, flowcell_name: str) -> StatsFlowcell:
        """Fetch information about a flowcell."""
        flowcell_object: Flowcell = self.Flowcell.query.filter_by(
            flowcellname=flowcell_name
        ).first()
        flowcell_data = {
            "name": flowcell_object.flowcellname,
            "sequencer": flowcell_object.demux[0].datasource.machine,
            "sequencer_type": flowcell_object.hiseqtype,
            "date": flowcell_object.time,
            "samples": self.get_flowcell_samples(flowcell_object),
        }

        return StatsFlowcell(**flowcell_data)

    def flowcell_samples(self, flowcell_obj: Flowcell) -> Iterator[Sample]:
        """Fetch all the samples from a flowcell."""
        return self.Sample.query.join(Sample.unaligned, Unaligned.demux).filter(
            Demux.flowcell == flowcell_obj
        )

    def is_lane_pooled(self, flowcell_obj: Flowcell, lane: str) -> bool:
        """Check whether a lane is pooled or not."""
        query = (
            self.session.query(sqa.func.count(Unaligned.sample_id).label("sample_count"))
            .join(Unaligned.demux)
            .filter(Demux.flowcell == flowcell_obj)
            .filter(Unaligned.lane == lane)
        )
        return query.first().sample_count > 1

    def sample_reads(self, sample_obj: Sample) -> alchy.Query:
        """Calculate reads for a sample."""
        return (
            self.session.query(
                Flowcell.flowcellname.label("name"),
                Flowcell.hiseqtype.label("type"),
                Unaligned.lane,
                Demux.basemask.label("base_mask"),
                sqa.func.sum(Unaligned.readcounts).label("reads"),
                sqa.func.min(Unaligned.q30_bases_pct).label("q30"),
            )
            .join(Flowcell.demux, Demux.unaligned)
            .filter(Unaligned.sample == sample_obj)
            .group_by(Flowcell.flowcellname)
        )

    def flow_cell_reads_and_q30_summary(self, flow_cell_name: str) -> Dict[str, Union[int, float]]:
        """Calculate reads and q30 for a flow cell."""
        flow_cell_reads_and_q30_summary: Dict[str, Union[int, float]] = {"reads": 0, "q30": 0.0}
        flow_cell_obj: Flowcell = self.find_handler.get_flow_cell_by_name(
            flow_cell_name=flow_cell_name
        )

        if flow_cell_obj:
            sample_count: int = 0
            q30_list: List[float] = []

            for sample in self.flowcell_samples(flowcell_obj=flow_cell_obj):
                sample_count += 1
                sample_info = self.sample_reads(sample_obj=sample).first()
                flow_cell_reads_and_q30_summary["reads"] += int(sample_info.reads)
                q30_list.append(float(sample_info.q30))

            flow_cell_reads_and_q30_summary["q30"]: float = sum(q30_list) / sample_count
        else:
            LOG.error(f"StatsAPI: Could not find flowcell in database with name: {flow_cell_name}")

        return flow_cell_reads_and_q30_summary

    def fastqs(self, flowcell: str, sample_obj: Sample) -> Iterator[Path]:
        """Fetch FASTQ files for a sample."""
        base_pattern = "*{}/Unaligned*/Project_*/Sample_{}/*.fastq.gz"
        alt_pattern = "*{}/Unaligned*/Project_*/Sample_{}_*/*.fastq.gz"
        for fastq_pattern in (base_pattern, alt_pattern):
            pattern = fastq_pattern.format(flowcell, sample_obj.samplename)
            yield from self.root_dir.glob(pattern)



import logging
from typing import List
from typing_extensions import Literal

from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.lims.samplesheet import LimsFlowcellSample
from cg.constants.sequencing import Sequencers
from cg.constants.demultiplexing import BclConverter
from cg.exc import FlowCellError
from cg.models.demultiplex.flow_cell import FlowCell

LOG = logging.getLogger(__name__)


def create_sample_sheet(
    bcl_converter: Literal[BclConverter.BCL2FASTQ, BclConverter.DRAGEN],
    flow_cell: FlowCell,
    lims_samples: List[LimsFlowcellSample],
    force: bool = False,
) -> List[List[str]]:
    """Create a sample sheet for a flow cell."""
    if flow_cell.sample_sheet_path.exists():
        message = f"Sample sheet {flow_cell.sample_sheet_path} already exists!"
        LOG.warning(message)
        raise FileExistsError(message)

    flow_cell_sequencer: str = flow_cell.sequencer_type

    if flow_cell_sequencer not in [Sequencers.NOVASEQ, Sequencers.NOVASEQX]:
        message = f"Only demultiplexing of Novaseq sequence data is currently supported. Found sequencer type: {flow_cell_sequencer}"
        LOG.warning(message)
        raise FlowCellError(message=message)

    sample_sheet_creator = SampleSheetCreator(
        bcl_converter=bcl_converter,
        flow_cell=flow_cell,
        lims_samples=lims_samples,
        force=force,
    )
    return sample_sheet_creator.construct_sample_sheet()



from cg.apps.cgstats.crud import create, find
from cg.apps.cgstats.db.models import Supportparams, Datasource, Flowcell
from cg.apps.cgstats.stats import StatsAPI
from cg.models.demultiplex.demux_results import DemuxResults


def test_create_support_parameters(stats_api: StatsAPI, bcl2fastq_demux_results: DemuxResults):
    # GIVEN a cg stats api without any support parameter for the demux result
    document_path = str(bcl2fastq_demux_results.results_dir)
    assert not stats_api.find_handler.get_support_parameters_by_document_path(
        document_path=document_path
    )

    # WHEN creating the new support parameters
    create.create_support_parameters(manager=stats_api, demux_results=bcl2fastq_demux_results)

    # THEN assert that the parameters was created
    document_path: str = str(bcl2fastq_demux_results.results_dir)
    support_parameters = stats_api.find_handler.get_support_parameters_by_document_path(
        document_path=document_path
    )
    assert isinstance(support_parameters, Supportparams)


def test_create_data_source(stats_api: StatsAPI, bcl2fastq_demux_results: DemuxResults):
    # GIVEN a api with some support parameters
    support_parameters: Supportparams = create.create_support_parameters(
        manager=stats_api, demux_results=bcl2fastq_demux_results
    )
    document_path: str = str(bcl2fastq_demux_results.conversion_stats_path)

    # GIVEN that there are no data source for the given run
    assert not stats_api.find_handler.get_datasource_by_document_path(document_path=document_path)

    # WHEN creating a new datasource
    create.create_datasource(
        manager=stats_api,
        demux_results=bcl2fastq_demux_results,
        support_parameters_id=support_parameters.supportparams_id,
    )

    # THEN assert that the datasource exists
    assert stats_api.find_handler.get_datasource_by_document_path(document_path=document_path)


def test_create_flowcell(stats_api: StatsAPI, bcl2fastq_demux_results: DemuxResults):
    # GIVEN a api without a flowcell object
    assert not stats_api.find_handler.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_demux_results.flow_cell.id
    )

    # WHEN creating a new flowcell
    create.create_flowcell(manager=stats_api, demux_results=bcl2fastq_demux_results)

    # THEN assert that the flowcell was created
    assert stats_api.find_handler.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_demux_results.flow_cell.id
    )


def test_create_demux(stats_api: StatsAPI, bcl2fastq_demux_results: DemuxResults):
    # GIVEN a database with a flowcell and a data source
    support_parameters: Supportparams = create.create_support_parameters(
        manager=stats_api, demux_results=bcl2fastq_demux_results
    )
    flowcell: Flowcell = create.create_flowcell(
        manager=stats_api, demux_results=bcl2fastq_demux_results
    )
    data_source: Datasource = create.create_datasource(
        manager=stats_api,
        demux_results=bcl2fastq_demux_results,
        support_parameters_id=support_parameters.supportparams_id,
    )
    # GIVEN that there is not demux object in the database
    assert not stats_api.find_handler.get_demux_by_flow_cell_id_and_base_mask(
        flowcell_id=flowcell.flowcell_id
    )

    # WHEN creating a demux object
    create.create_demux(
        manager=stats_api,
        demux_results=bcl2fastq_demux_results,
        flow_cell_id=flowcell.flowcell_id,
        datasource_id=data_source.datasource_id,
    )

    # THEN assert that a demux object was created
    assert stats_api.find_handler.get_demux_by_flow_cell_id_and_base_mask(
        flowcell_id=flowcell.flowcell_id
    )


def test_create_dragen_demux(stats_api: StatsAPI, dragen_demux_results: DemuxResults):
    # GIVEN a database with a flowcell and a data source
    support_parameters: Supportparams = create.create_support_parameters(
        manager=stats_api, demux_results=dragen_demux_results
    )
    flowcell: Flowcell = create.create_flowcell(
        manager=stats_api, demux_results=dragen_demux_results
    )
    data_source: Datasource = create.create_datasource(
        manager=stats_api,
        demux_results=dragen_demux_results,
        support_parameters_id=support_parameters.supportparams_id,
    )
    # GIVEN that there is not demux object in the database
    assert not stats_api.find_handler.get_demux_by_flow_cell_id_and_base_mask(
        flowcell_id=flowcell.flowcell_id
    )

    # WHEN creating a demux object
    demux_object = create.create_demux(
        manager=stats_api,
        demux_results=dragen_demux_results,
        flow_cell_id=flowcell.flowcell_id,
        datasource_id=data_source.datasource_id,
    )

    # THEN assert that a demux object was created
    assert stats_api.find_handler.get_demux_by_flow_cell_id_and_base_mask(
        flowcell_id=flowcell.flowcell_id, base_mask=demux_object.basemask
    )



"""Tests the find business data part of the Cg store API related to sample model."""

import pytest
from typing import List, Dict, Any
from sqlalchemy.orm import Query
from cg.store import Store

from cg.store.models import Sample, Invoice, Customer
from tests.store_helpers import StoreHelpers


def test_get_all_pools_and_samples_for_invoice_by_invoice_id(store: Store, helpers: StoreHelpers):
    """Test that all pools and samples for an invoice can be fetched."""

    # GIVEN a database with a pool and a sample
    pool = helpers.ensure_pool(store=store, name="pool_1")
    sample = helpers.add_sample(store=store, name="sample_1")

    # AND an invoice with the pool and sample
    invoice: Invoice = helpers.ensure_invoice(store=store, pools=[pool], samples=[sample])

    # ASSERT that there is an invoice with a pool and a sample
    assert len(invoice.pools) == 1
    assert len(invoice.samples) == 1

    # WHEN fetching all pools and samples for the invoice
    records = store.get_pools_and_samples_for_invoice_by_invoice_id(invoice_id=invoice.id)
    # THEN the pool and sample should be returned
    assert pool in records
    assert sample in records


def test_get_samples_by_customer_and_subject_id_query(
    store_with_samples_subject_id_and_tumour_status: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # GIVEN that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status._get_query(table=Sample).all()) == 2

    # GIVEN that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_internal_id(
        customer_internal_id=cust123
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = store_with_samples_subject_id_and_tumour_status._get_samples_by_customer_and_subject_id_query(
        subject_id=test_subject, customer_internal_id=cust123
    )

    # THEN two samples should be returned
    assert samples.count() == 2
    # THEN the fetched samples have the subject id used for filtering
    assert all(fetched_sample.subject_id == test_subject for fetched_sample in samples.all())


def test_get_samples_by_customer_and_subject_id_query_missing_subject_id(
    store_with_samples_and_tumour_status_missing_subject_id: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # GIVEN that there are two samples in the store
    assert (
        len(store_with_samples_and_tumour_status_missing_subject_id._get_query(table=Sample).all())
        == 2
    )

    # GIVEN that there is a customer with the given customer id
    assert store_with_samples_and_tumour_status_missing_subject_id.get_customer_by_internal_id(
        customer_internal_id=cust123
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = store_with_samples_and_tumour_status_missing_subject_id._get_samples_by_customer_and_subject_id_query(
        subject_id=test_subject, customer_internal_id=cust123
    )

    # THEN two samples should be returned
    assert samples.count() == 0


def test_get_samples_by_subject_id(
    store_with_samples_subject_id_and_tumour_status: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status._get_query(table=Sample).all()) == 2

    # ASSERT that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_internal_id(
        customer_internal_id=cust123
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = (
        store_with_samples_subject_id_and_tumour_status.get_samples_by_customer_and_subject_id(
            subject_id=test_subject, customer_internal_id=cust123
        )
    )

    # THEN two samples should be returned
    assert samples and len(samples) == 2


def test_get_samples_by_subject_id_and_is_tumour(
    store_with_samples_subject_id_and_tumour_status: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
    is_tumour: bool = True,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status._get_query(table=Sample).all()) == 2

    # ASSERT that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_internal_id(
        customer_internal_id=cust123
    )
    # WHEN fetching the sample by subject id and customer_id
    samples: List[
        Sample
    ] = store_with_samples_subject_id_and_tumour_status.get_samples_by_customer_subject_id_and_is_tumour(
        subject_id=test_subject, customer_internal_id=cust123, is_tumour=is_tumour
    )

    # THEN two samples should be returned
    assert samples and len(samples) == 1


def test_get_samples_by_customer_id_list_and_subject_id_and_is_tumour(
    store_with_samples_customer_id_and_subject_id_and_tumour_status: Store,
    customer_ids: List[int] = [1, 2],
    subject_id: str = "test_subject",
    is_tumour: bool = True,
):
    """Test that samples can be fetched by customer ID, subject ID, and tumour status."""
    # GIVEN a database with four samples, two with customer ID 1 and two with customer ID 2

    # ASSERT that there are customers with the given customer IDs
    for customer_id in customer_ids:
        assert store_with_samples_customer_id_and_subject_id_and_tumour_status.get_customer_by_internal_id(
            customer_internal_id=str(customer_id)
        )

    # WHEN fetching the samples by customer ID list, subject ID, and tumour status
    samples = store_with_samples_customer_id_and_subject_id_and_tumour_status.get_samples_by_customer_id_list_and_subject_id_and_is_tumour(
        customer_ids=customer_ids, subject_id=subject_id, is_tumour=is_tumour
    )

    # THEN two samples should be returned, one for each customer ID, with the specified subject ID and tumour status
    assert isinstance(samples, list)
    assert len(samples) == 2

    for sample in samples:
        assert isinstance(sample, Sample)

    for customer_id, sample in zip(customer_ids, samples):
        assert sample.customer_id == customer_id
        assert sample.subject_id == subject_id
        assert sample.is_tumour == is_tumour


def test_get_samples_by_customer_id_list_and_subject_id_and_is_tumour_with_non_existing_customer_id(
    store_with_samples_customer_id_and_subject_id_and_tumour_status: Store,
    helpers: StoreHelpers,
):
    """Test that no samples are returned when filtering on non-existing customer ID."""
    # GIVEN a database with four samples, two with customer ID 1 and two with customer ID 2

    # ASSERT that there are no customers with the given customer IDs
    customer_ids = [1, 2, 3]
    for customer_id in customer_ids:
        if customer_id == 3:
            assert (
                store_with_samples_customer_id_and_subject_id_and_tumour_status.get_customer_by_internal_id(
                    customer_internal_id=str(customer_id)
                )
                is None
            )
        else:
            assert store_with_samples_customer_id_and_subject_id_and_tumour_status.get_customer_by_internal_id(
                customer_internal_id=str(customer_id)
            )

    # WHEN fetching the samples by customer ID list, subject ID, and tumour status
    non_existing_customer_id = [3]
    samples = store_with_samples_customer_id_and_subject_id_and_tumour_status.get_samples_by_customer_id_list_and_subject_id_and_is_tumour(
        customer_ids=non_existing_customer_id, subject_id="test_subject", is_tumour=True
    )

    # THEN no samples should be returned
    assert isinstance(samples, list)
    assert len(samples) == 0


def test_get_sample_by_name(store_with_samples_that_have_names: Store, name="test_sample_1"):
    """Test that samples can be fetched by name."""
    # GIVEN a database with two samples of which one has a name

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_that_have_names._get_query(table=Sample).all()) == 4

    # WHEN fetching the sample by name
    samples: Sample = store_with_samples_that_have_names.get_sample_by_name(name=name)

    # THEN one sample should be returned
    assert samples and samples.name == name


def test_get_samples_by_name_pattern(
    store_with_samples_that_have_names: Store, name_pattern="sample"
):
    """Test that samples can be fetched by name."""
    # GIVEN a database with two samples of which one has a name

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_that_have_names._get_query(table=Sample).all()) == 4

    # WHEN fetching the sample by name
    samples: List[Sample] = store_with_samples_that_have_names.get_samples_by_name_pattern(
        name_pattern=name_pattern
    )

    # THEN one sample should be returned
    assert samples and len(samples) == 3


def test_has_active_cases_for_sample_analyze(
    store_with_active_sample_analyze: Store,
    helpers: StoreHelpers,
    sample_internal_id: str = "test_sample_internal_id",
):
    """Test that a sample is active."""
    # GIVEN a store with an active case

    # THEN the sample should be active
    assert (
        store_with_active_sample_analyze.has_active_cases_for_sample(internal_id=sample_internal_id)
        is True
    )


def test_has_active_cases_for_sample_running(
    store_with_active_sample_running: Store,
    helpers: StoreHelpers,
    sample_internal_id: str = "test_sample_internal_id",
):
    """Test that a sample is active."""
    # GIVEN a store with an active case

    # THEN the sample should be active
    assert (
        store_with_active_sample_running.has_active_cases_for_sample(internal_id=sample_internal_id)
        is True
    )


def test_get_samples_by_customer_and_name(
    store_with_samples_that_have_names: Store,
    helpers: StoreHelpers,
    name: str = "test_sample_1",
    customer_id="cust000",
):
    """Test that samples can be fetched by name."""
    # GIVEN a database with samples
    # WHEN getting a customer from the store
    customer: Customer = store_with_samples_that_have_names.get_customer_by_internal_id(
        customer_internal_id=customer_id
    )
    assert customer
    # WHEN fetching the sample by name
    sample: Sample = store_with_samples_that_have_names.get_sample_by_customer_and_name(
        customer_entry_id=[customer.id], sample_name=name
    )

    # THEN one sample should be returned
    assert sample
    assert sample.name == name
    assert sample.customer == customer


def test_get_samples_by_customer_and_name_invalid_customer(
    store_with_samples_that_have_names: Store,
    helpers: StoreHelpers,
    name: str = "test_sample_1",
    customer_id="unrelated_customer",
):
    """Test that samples can be fetched by name does not return a sample when invalid customer is supplied."""
    # GIVEN a database with two samples

    # WHEN getting a customer from the store
    customer: Customer = store_with_samples_that_have_names.get_customer_by_internal_id(
        customer_internal_id=customer_id
    )
    assert customer
    # WHEN fetching the sample by name
    sample: Sample = store_with_samples_that_have_names.get_sample_by_customer_and_name(
        customer_entry_id=[customer.id], sample_name=name
    )

    # THEN one sample should be returned
    assert not sample


def test_get_samples_by_any_id_not_an_attribute_fails(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    identifiers: Dict[str, Any] = {
        "non-existent-attribute": "not-an-attribute",
        "non-existent-value": "not-a-value",
    },
):
    """Test that using an attribute not from Sample raises an error."""
    # GIVEN a store with samples
    sample_query: Query = store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
        table=Sample
    )
    assert sample_query.count() > 0

    # WHEN trying to filter using an attribute not of Sample
    with pytest.raises(AttributeError) as exc_info:
        store_with_a_sample_that_has_many_attributes_and_one_without.get_samples_by_any_id(
            **identifiers
        )

    # THEN the error message should contain the non-existent-attribute
    assert str(exc_info.value) == "type object 'Sample' has no attribute 'non-existent-attribute'"


def test_get_samples_by_any_id_exclusive_filtering_gives_empty_query(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that using mutually exclusive filtering conditions give an empty query."""
    # GIVEN a store with two samples
    sample_query: Query = store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
        table=Sample
    )
    assert sample_query.count() == 2

    # GIVEN that the samples in the query have different values of two attributes
    sample_1: Sample = sample_query[0]
    sample_2: Sample = sample_query[1]
    assert sample_1.name != sample_2.name
    assert sample_1.is_tumour != sample_2.is_tumour

    # WHEN filtering twice mutually exclusive conditions
    identifiers: Dict[str, str] = {
        "name": sample_1.name,
        "is_tumour": sample_2.is_tumour,
    }
    filtered_query: Query = (
        store_with_a_sample_that_has_many_attributes_and_one_without.get_samples_by_any_id(
            **identifiers
        )
    )

    # THEN the filtered query is empty
    assert filtered_query.count() == 0



"""Tests the cli for generating delivery reports"""

import logging
from datetime import datetime

from cg.cli.generate.report.base import generate_delivery_report
from cg.constants import EXIT_SUCCESS, EXIT_FAIL


def test_delivery_report_invalid_case(mip_dna_context, cli_runner, caplog):
    """Tests the delivery report command for an invalid case"""

    caplog.set_level(logging.INFO)

    # GIVEN a MIP DNA context object

    # WHEN calling delivery_report with an invalid case name
    result = cli_runner.invoke(
        generate_delivery_report,
        ["not a case"],
        obj=mip_dna_context,
    )

    # THEN the command should fail due to an invalid case ID
    assert "Invalid case ID. Retrieving available cases." in caplog.text
    assert "There are no valid cases to perform delivery report actions" in result.output
    assert result.exit_code == EXIT_FAIL


def test_delivery_report_dry_run(mip_dna_context, cli_runner, case_id, caplog):
    """Tests the delivery report command with a dry run option"""

    caplog.set_level(logging.INFO)

    # GIVEN a MIP DNA context object

    # WHEN calling delivery_report with a dry option
    result = cli_runner.invoke(
        generate_delivery_report,
        [case_id, "--dry-run"],
        obj=mip_dna_context,
    )

    # THEN the command should be invoked successfully
    assert "create_delivery_report" in caplog.text
    assert "create_delivery_report_file" not in caplog.text
    assert "case=" + case_id in caplog.text
    assert "analysis_date=" + str(datetime.now().date()) in caplog.text
    assert result.exit_code == EXIT_SUCCESS


def test_delivery_report(mip_dna_context, cli_runner, case_id, caplog):
    """Tests the delivery report command expecting a rendered html file"""

    caplog.set_level(logging.INFO)

    # GIVEN a MIP DNA context object

    # WHEN calling delivery_report with a dry option
    cli_runner.invoke(
        generate_delivery_report,
        [case_id],
        obj=mip_dna_context,
    )

    # THEN the report html should have been created and added to HK
    assert "create_delivery_report_file" in caplog.text
    assert "add_delivery_report_to_hk" in caplog.text



"""This module parses metrics for files generated by the BCLConvert tool using the Dragen hardware."""
import csv
import xml.etree.ElementTree as ET
import logging
from pathlib import Path
from typing import List, Union, Dict, Callable
from cg.io.controller import ReadFile
from cg.constants.demultiplexing import SampleSheetHeaderColumnNames
from cg.constants.constants import FileFormat
from cg.apps.sequencing_metrics_parser.models.bcl_convert import (
    BclConvertAdapterMetrics,
    BclConvertDemuxMetrics,
    BclConvertQualityMetrics,
    BclConvertRunInfo,
    BclConvertSampleSheetData,
)


LOG = logging.getLogger(__name__)


class BclConvertMetricsParser:
    def __init__(
        self,
        bcl_convert_quality_metrics_file_path: Path,
        bcl_convert_demux_metrics_file_path: Path,
        bcl_convert_adapter_metrics_file_path: Path,
        bcl_convert_sample_sheet_file_path: Path,
        bcl_convert_run_info_file_path: Path,
    ) -> None:
        """Initialize the class."""
        self.quality_metrics_path: Path = bcl_convert_quality_metrics_file_path
        self.demux_metrics_path: Path = bcl_convert_demux_metrics_file_path
        self.adapter_metrics_path: Path = bcl_convert_adapter_metrics_file_path
        self.sample_sheet_path: Path = bcl_convert_sample_sheet_file_path
        self.run_info_path: Path = bcl_convert_run_info_file_path
        self.quality_metrics: List[BclConvertQualityMetrics] = self.parse_metrics_file(
            metrics_file_path=self.quality_metrics_path, metrics_model=BclConvertQualityMetrics
        )
        self.demux_metrics: List[BclConvertDemuxMetrics] = self.parse_metrics_file(
            metrics_file_path=self.demux_metrics_path, metrics_model=BclConvertDemuxMetrics
        )
        self.adapter_metrics: List[BclConvertAdapterMetrics] = self.parse_metrics_file(
            metrics_file_path=self.adapter_metrics_path, metrics_model=BclConvertAdapterMetrics
        )
        self.sample_sheet: List[BclConvertSampleSheetData] = self.parse_sample_sheet_file()
        self.run_info: BclConvertRunInfo = self.parse_run_info_file()

    def parse_metrics_file(
        self, metrics_file_path, metrics_model: Callable
    ) -> List[Union[BclConvertQualityMetrics, BclConvertDemuxMetrics, BclConvertAdapterMetrics]]:
        """Parse specified BCL convert metrics file."""
        LOG.info(f"Parsing BCLConvert metrics file: {metrics_file_path}")
        parsed_metrics: List[
            Union[BclConvertQualityMetrics, BclConvertDemuxMetrics, BclConvertAdapterMetrics]
        ] = []
        metrics_content: List[Dict] = ReadFile.get_content_from_file(
            file_format=FileFormat.CSV, file_path=metrics_file_path, read_to_dict=True
        )
        for sample_metrics_dict in metrics_content:
            parsed_metrics.append(metrics_model(**sample_metrics_dict))
        return parsed_metrics

    def get_nr_of_header_lines_in_sample_sheet(
        self,
    ) -> int:
        """Return the number of header lines in a sample sheet.
        Any lines before and including the line starting with [Data] is considered the header."""
        sample_sheet_content = ReadFile.get_content_from_file(
            FileFormat.CSV, self.sample_sheet_path
        )
        header_line_count: int = 1
        for line in sample_sheet_content:
            if SampleSheetHeaderColumnNames.DATA.value in line:
                break
            header_line_count += 1
        return header_line_count

    def parse_sample_sheet_file(self) -> List[BclConvertSampleSheetData]:
        """Return sample sheet sample lines."""
        LOG.info(f"Parsing BCLConvert sample sheet file: {self.sample_sheet_path}")
        header_line_count: int = self.get_nr_of_header_lines_in_sample_sheet()
        sample_sheet_sample_lines: List[BclConvertSampleSheetData] = []
        with open(self.sample_sheet_path, "r") as sample_sheet_file:
            for _ in range(header_line_count):
                next(sample_sheet_file)
            sample_sheet_content = csv.DictReader(sample_sheet_file)
            for line in sample_sheet_content:
                sample_sheet_sample_lines.append(BclConvertSampleSheetData(**line))
        return sample_sheet_sample_lines

    def parse_run_info_file(self) -> BclConvertRunInfo:
        LOG.info(f"Parsing Run info XML {self.run_info_path}")
        parsed_metrics = BclConvertRunInfo(tree=ET.parse(self.run_info_path))
        return parsed_metrics



"""Functions that handle files in the context of scout uploading"""
import logging
from typing import List, Optional, Set

import requests
from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.models.scout.scout_load_config import ScoutIndividual, ScoutLoadConfig
from cg.store.models import Analysis, FamilySample, Sample

LOG = logging.getLogger(__name__)

# Maps keys that are used in scout load config on tags that are used in scout


class ScoutConfigBuilder:
    """Base class for handling files that should be included in Scout upload"""

    def __init__(self, hk_version_obj: Version, analysis_obj: Analysis, lims_api: LimsAPI):
        self.hk_version_obj: Version = hk_version_obj
        self.analysis_obj: Analysis = analysis_obj
        self.lims_api: LimsAPI = lims_api
        self.case_tags: CaseTags
        self.sample_tags: SampleTags
        self.load_config: ScoutLoadConfig = ScoutLoadConfig()

    def add_common_info_to_load_config(self) -> None:
        """Add the mandatory common information to a scout load config object"""
        self.load_config.analysis_date = self.analysis_obj.completed_at
        self.load_config.default_gene_panels = self.analysis_obj.family.panels
        self.load_config.family = self.analysis_obj.family.internal_id
        self.load_config.family_name = self.analysis_obj.family.name
        self.load_config.owner = self.analysis_obj.family.customer.internal_id
        self.load_config.synopsis = self.analysis_obj.family.synopsis
        self.include_cohorts()
        self.include_phenotype_groups()
        self.include_phenotype_terms()

    def add_common_sample_info(
        self, config_sample: ScoutIndividual, case_sample: FamilySample
    ) -> None:
        """Add the information to a sample that is common for different analysis types"""
        sample_id: str = case_sample.sample.internal_id
        LOG.info("Building sample %s", sample_id)
        lims_sample = dict()
        try:
            lims_sample = self.lims_api.sample(sample_id) or {}
        except requests.exceptions.HTTPError as ex:
            LOG.info("Could not fetch sample %s from LIMS: %s", sample_id, ex)

        config_sample.sample_id = sample_id
        config_sample.sex = case_sample.sample.sex
        config_sample.phenotype = case_sample.status
        config_sample.analysis_type = (
            case_sample.sample.application_version.application.analysis_type
        )
        config_sample.sample_name = case_sample.sample.name
        config_sample.tissue_type = lims_sample.get("source", "unknown")
        config_sample.subject_id = case_sample.sample.subject_id

    def add_common_sample_files(
        self,
        config_sample: ScoutIndividual,
        case_sample: FamilySample,
    ) -> None:
        """Add common sample files for different analysis types."""
        sample_id: str = case_sample.sample.internal_id
        LOG.info(f"Adding common files for sample {sample_id}")
        self.include_sample_alignment_file(config_sample=config_sample)
        self.include_sample_files(config_sample=config_sample)

    def build_config_sample(self, case_sample: FamilySample) -> ScoutIndividual:
        """Build a sample for the scout load config"""
        raise NotImplementedError

    def build_load_config(self) -> ScoutLoadConfig:
        """Build a load config for uploading a case to scout"""
        raise NotImplementedError

    def include_sample_files(self, config_sample: ScoutIndividual) -> None:
        """Include all files that are used on sample level in Scout"""
        raise NotImplementedError

    def include_case_files(self) -> None:
        """Include all files that are used on case level in scout"""
        raise NotImplementedError

    def include_phenotype_terms(self) -> None:
        LOG.info("Adding phenotype terms to scout load config")
        phenotype_terms: Set[str] = set()
        link_obj: FamilySample
        for link_obj in self.analysis_obj.family.links:
            sample_obj: Sample = link_obj.sample
            for phenotype_term in sample_obj.phenotype_terms:
                LOG.debug(
                    "Adding term %s from sample %s to phenotype terms",
                    phenotype_term,
                    sample_obj.internal_id,
                )
                phenotype_terms.add(phenotype_term)
        if phenotype_terms:
            self.load_config.phenotype_terms = list(phenotype_terms)

    def include_phenotype_groups(self) -> None:
        LOG.info("Adding phenotype groups to scout load config")
        phenotype_groups: Set[str] = set()
        link_obj: FamilySample
        for link_obj in self.analysis_obj.family.links:
            sample_obj: Sample = link_obj.sample
            for phenotype_group in sample_obj.phenotype_groups:
                LOG.debug(
                    "Adding group %s from sample %s to phenotype groups",
                    phenotype_group,
                    sample_obj.internal_id,
                )
                phenotype_groups.add(phenotype_group)
        if phenotype_groups:
            self.load_config.phenotype_groups = list(phenotype_groups)

    def include_synopsis(self) -> None:
        LOG.info("Adding synopsis string to scout load config")
        synopsis_string: str = self.analysis_obj.family.synopsis
        if synopsis_string:
            LOG.debug("Adding synopsis string %s", synopsis_string)
            self.load_config.synopsis = synopsis_string

    def include_cohorts(self) -> None:
        LOG.info("Including cohorts to scout load config")
        cohorts: List[str] = self.analysis_obj.family.cohorts
        if cohorts:
            LOG.debug("Adding cohorts %s", ", ".join(cohorts))
            self.load_config.cohorts = cohorts

    def include_cnv_report(self) -> None:
        LOG.info("Include CNV report to case")
        self.load_config.cnv_report = self.get_file_from_hk(
            hk_tags=self.case_tags.cnv_report, latest=True
        )

    def include_multiqc_report(self) -> None:
        LOG.info("Include MultiQC report to case")
        self.load_config.multiqc = self.get_file_from_hk(
            hk_tags=self.case_tags.multiqc_report, latest=True
        )

    def include_delivery_report(self) -> None:
        LOG.info("Include delivery report to case")
        self.load_config.delivery_report = self.get_file_from_hk(
            hk_tags=self.case_tags.delivery_report, latest=True
        )

    def include_sample_alignment_file(self, config_sample: ScoutIndividual) -> None:
        """Include the alignment file for a sample

        First add the bam.
        Cram is preferred so overwrite if found
        """
        sample_id: str = config_sample.sample_id
        config_sample.alignment_path = self.get_sample_file(
            hk_tags=self.sample_tags.bam_file, sample_id=sample_id
        )

        config_sample.alignment_path = self.get_sample_file(
            hk_tags=self.sample_tags.alignment_file, sample_id=sample_id
        )

    def get_sample_file(self, hk_tags: Set[str], sample_id: str) -> Optional[str]:
        """Return a file that is specific for a individual from housekeeper"""
        tags: set = hk_tags.copy()
        tags.add(sample_id)
        return self.get_file_from_hk(hk_tags=tags)

    def get_file_from_hk(self, hk_tags: Set[str], latest: Optional[bool] = False) -> Optional[str]:
        """Get a file from housekeeper and return the path as a string."""
        LOG.info(f"Get file with tags {hk_tags}")
        if not hk_tags:
            LOG.debug("No tags provided, skipping")
            return None
        hk_file: Optional[File] = (
            HousekeeperAPI.get_latest_file_from_version(version=self.hk_version_obj, tags=hk_tags)
            if latest
            else HousekeeperAPI.get_file_from_version(version=self.hk_version_obj, tags=hk_tags)
        )
        return hk_file.full_path if hk_file else None



"""Tests for the demultiplex flow cell class."""
from pathlib import Path

import pytest
from cg.models.demultiplex.flow_cell import FlowCell


def test_get_run_parameters_when_non_existing(fixtures_dir: Path):
    # GIVEN a flowcell object with a directory without run parameters
    flowcell_path: Path = (
        fixtures_dir
        / "apps"
        / "demultiplexing"
        / "demultiplexed-runs"
        / "201203_A00689_0200_AHVKJCDRXX"
    )
    flow_cell = FlowCell(flow_cell_path=flowcell_path)
    assert flow_cell.run_parameters_path.exists() is False

    # WHEN fetching the run parameters object
    with pytest.raises(FileNotFoundError):
        # THEN assert that a FileNotFound error is raised
        flow_cell.run_parameters



"""CLI support for tower pipelines."""

import click

OPTION_COMPUTE_ENV = click.option(
    "--compute-env",
    type=str,
    help="Compute environment name. If not specified the primary compute environment will be used.",
)



from pydantic import BaseModel, BaseConfig, Field
import xml.etree.ElementTree as ET
from cg.constants.bcl_convert_metrics import (
    BclConvertQualityMetricsColumnNames,
    BclConvertDemuxMetricsColumnNames,
    BclConvertAdapterMetricsColumnNames,
)
from cg.constants.demultiplexing import SampleSheetHeaderColumnNames


class BclConvertQualityMetrics(BaseModel):
    """Model for the BCL Convert quality metrics."""

    lane: int = Field(..., alias=BclConvertQualityMetricsColumnNames.LANE.value)
    sample_internal_id: str = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.SAMPLE_INTERNAL_ID.value
    )
    read_pair_number: int = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.READ_PAIR_NUMBER.value
    )
    yield_bases: int = Field(..., alias=BclConvertQualityMetricsColumnNames.YIELD_BASES.value)
    yield_q30_bases: int = Field(..., alias=BclConvertQualityMetricsColumnNames.YIELD_Q30.value)
    quality_score_sum: int = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.QUALITY_SCORE_SUM.value
    )
    mean_quality_score_q30: float = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.MEAN_QUALITY_SCORE_Q30.value
    )
    q30_bases_percent: float = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.Q30_BASES_PERCENT.value
    )


class BclConvertDemuxMetrics(BaseModel):
    """Model for the BCL Convert demultiplexing metrics."""

    lane: int = Field(..., alias=BclConvertDemuxMetricsColumnNames.LANE.value)
    sample_internal_id: str = Field(
        ..., alias=BclConvertDemuxMetricsColumnNames.SAMPLE_INTERNAL_ID.value
    )
    sample_project: str = Field(..., alias=BclConvertDemuxMetricsColumnNames.SAMPLE_PROJECT.value)
    read_pair_count: int = Field(..., alias=BclConvertDemuxMetricsColumnNames.READ_PAIR_COUNT.value)
    perfect_index_reads_count: int = Field(
        ..., alias=BclConvertDemuxMetricsColumnNames.PERFECT_INDEX_READS_COUNT.value
    )
    perfect_index_reads_percent: float = Field(
        ..., alias=BclConvertDemuxMetricsColumnNames.PERFECT_INDEX_READS_PERCENT.value
    )
    one_mismatch_index_reads_count: int = Field(
        ..., alias=BclConvertDemuxMetricsColumnNames.ONE_MISMATCH_INDEX_READS_COUNT.value
    )
    two_mismatch_index_reads_count: int = Field(
        ..., alias=BclConvertDemuxMetricsColumnNames.TWO_MISMATCH_INDEX_READS_COUNT.value
    )


class BclConvertAdapterMetrics(BaseModel):
    """Model for the BCL Convert adapter metrics."""

    lane: int = Field(..., alias=BclConvertAdapterMetricsColumnNames.LANE.value)
    sample_internal_id: str = Field(
        ..., alias=BclConvertAdapterMetricsColumnNames.SAMPLE_INTERNAL_ID.value
    )
    sample_project: str = Field(..., alias=BclConvertAdapterMetricsColumnNames.SAMPLE_PROJECT.value)
    read_number: int = Field(..., alias=BclConvertAdapterMetricsColumnNames.READ_NUMBER.value)
    sample_bases: int = Field(..., alias=BclConvertAdapterMetricsColumnNames.SAMPLE_BASES.value)


class BclConvertSampleSheetData(BaseModel):
    """Model for the BCL Convert sample sheet."""

    flow_cell_name: str = Field(..., alias=SampleSheetHeaderColumnNames.FLOW_CELL_ID.value)
    lane: int = Field(..., alias=SampleSheetHeaderColumnNames.LANE.value)
    sample_internal_id: str = Field(
        ..., alias=SampleSheetHeaderColumnNames.SAMPLE_INTERNAL_ID.value
    )
    sample_name: str = Field(..., alias=SampleSheetHeaderColumnNames.SAMPLE_NAME.value)
    control: str = Field(..., alias=SampleSheetHeaderColumnNames.CONTROL.value)
    sample_project: str = Field(..., alias=SampleSheetHeaderColumnNames.SAMPLE_PROJECT.value)


class CustomConfig(BaseConfig):
    arbitrary_types_allowed = True


class BclConvertRunInfo(BaseModel):
    """Model for the BCL convert run info file."""

    tree: ET.ElementTree

    class Config(CustomConfig):
        pass
